"""Session routes."""

from __future__ import annotations

import asyncio
import contextlib
from typing import TYPE_CHECKING, Any

from anyenv.text_sharing.opencode import Message, MessagePart, OpenCodeSharer
from fastapi import APIRouter, HTTPException
from pydantic_ai import FileUrl

from agentpool.repomap import RepoMap, find_src_files
from agentpool.utils import identifiers as identifier
from agentpool.utils.time_utils import now_ms
from agentpool_server.opencode_server.command_validation import validate_command
from agentpool_server.opencode_server.converters import (
    chat_message_to_opencode,
    opencode_to_session_data,
    session_data_to_opencode,
)
from agentpool_server.opencode_server.dependencies import StateDep
from agentpool_server.opencode_server.input_provider import OpenCodeInputProvider
from agentpool_server.opencode_server.models import (
    AssistantMessage,
    CommandRequest,
    FileDiff,
    MessagePath,
    MessageTime,
    MessageUpdatedEvent,
    MessageWithParts,
    PartUpdatedEvent,
    Session,
    SessionCreatedEvent,
    SessionCreateRequest,
    SessionDeletedEvent,
    SessionForkRequest,
    SessionInitRequest,
    SessionRevert,
    SessionShare,
    SessionStatus,
    SessionStatusEvent,
    SessionUpdatedEvent,
    SessionUpdateRequest,
    ShellRequest,
    StepFinishPart,
    StepStartPart,
    SummarizeRequest,
    TextPart,
    TimeCreatedUpdated,
    Todo,
    Tokens,
)
from agentpool_server.opencode_server.models.base import OpenCodeBaseModel
from agentpool_server.opencode_server.models.events import (
    CommandExecutedEvent,
    PermissionAskedProperties,
    PermissionReplyRequest,
    PermissionResolvedEvent,
    SessionDiffEvent,
)


if TYPE_CHECKING:
    from agentpool_server.opencode_server.state import ServerState


async def get_or_load_session(state: ServerState, session_id: str) -> Session | None:
    """Get session from cache or load via agent.

    Returns None if session not found.
    Uses agent.load_session() which handles loading from the appropriate
    storage (pool storage, Claude storage, ACP server, Codex, etc.).
    """
    # Check if session AND messages are already loaded
    if session_id in state.sessions and session_id in state.messages:
        return state.sessions[session_id]

    # Load via agent - this populates agent.conversation.chat_messages
    data = await state.agent.load_session(session_id)
    if data is None:
        return None

    # Convert SessionData to OpenCode Session
    session = session_data_to_opencode(data)
    # Cache the session
    state.sessions[session_id] = session
    # Initialize runtime state
    if session_id not in state.session_status:
        state.session_status[session_id] = SessionStatus(type="idle")
    # Convert agent's conversation history to OpenCode format
    state.messages[session_id] = [
        chat_message_to_opencode(
            chat_msg,
            session_id=session_id,
            working_dir=state.working_dir,
            agent_name=state.agent.name,
            model_id=chat_msg.model_name or "sonnet",  # Normalized name from Claude storage
            provider_id=chat_msg.provider_name or "claude-code",
        )
        for chat_msg in state.agent.conversation.chat_messages
    ]
    return session


router = APIRouter(prefix="/session", tags=["session"])


@router.get("")
async def list_sessions(
    state: StateDep,
    roots: bool | None = None,
    start: int | None = None,
    search: str | None = None,
    limit: int | None = None,
) -> list[Session]:
    """List all sessions from the agent.

    Delegates to agent.list_sessions() which handles fetching sessions
    from the appropriate storage (pool storage, Claude storage, ACP server, etc.).

    Query params:
        roots: Only return root sessions (no parentID)
        start: Filter sessions updated on or after this timestamp (ms since epoch)
        search: Filter sessions by title (case-insensitive)
        limit: Maximum number of sessions to return
    """
    # Convert to OpenCode Session format and cache
    sessions: list[Session] = []
    for data in await state.agent.list_sessions(cwd=state.agent.env.cwd):
        session = session_data_to_opencode(data)
        # Cache in state for later use
        state.sessions[data.session_id] = session
        sessions.append(session)
    # Apply filters
    if roots:
        sessions = [s for s in sessions if s.parent_id is None]
    if start is not None:
        sessions = [s for s in sessions if s.time.updated >= start]
    if search:
        lower_search = search.lower()
        sessions = [s for s in sessions if lower_search in s.title.lower()]
    if limit is not None:
        sessions = sessions[:limit]
    return sessions


@router.post("")
async def create_session(state: StateDep, request: SessionCreateRequest | None = None) -> Session:
    """Create a new session and persist to storage."""
    now = now_ms()
    session_id = identifier.ascending("session")
    session = Session(
        id=session_id,
        project_id="default",  # TODO: Get from config/request
        directory=state.working_dir,
        title=request.title if request and request.title else "New Session",
        version="1",
        time=TimeCreatedUpdated(created=now, updated=now),
        parent_id=request.parent_id if request else None,
    )

    # Persist to storage
    id_ = state.pool.manifest.config_file_path
    session_data = opencode_to_session_data(session, agent_name=state.agent.name, pool_id=id_)
    await state.storage.save_session(session_data)
    # Cache in memory
    state.sessions[session_id] = session
    state.messages[session_id] = []
    state.session_status[session_id] = SessionStatus(type="idle")
    state.todos[session_id] = []
    # Create input provider for this session
    input_provider = OpenCodeInputProvider(state, session_id)
    state.input_providers[session_id] = input_provider
    # Set input provider on agent
    state.agent._input_provider = input_provider
    await state.broadcast_event(SessionCreatedEvent.create(session))
    return session


@router.get("/status")
async def get_session_status(state: StateDep) -> dict[str, SessionStatus]:
    """Get status for all sessions.

    Returns only non-idle sessions. If all sessions are idle, returns empty dict.
    """
    return {sid: status for sid, status in state.session_status.items() if status.type != "idle"}


@router.get("/{session_id}")
async def get_session(session_id: str, state: StateDep) -> Session:
    """Get session details.

    Loads from storage if not in memory cache.
    """
    session = await get_or_load_session(state, session_id)
    if session is None:
        raise HTTPException(status_code=404, detail="Session not found")
    return session


@router.patch("/{session_id}")
async def update_session(
    session_id: str,
    request: SessionUpdateRequest,
    state: StateDep,
) -> Session:
    """Update session properties and persist changes.

    Supports updating title and archiving via time.archived.
    """
    session = await get_or_load_session(state, session_id)
    if session is None:
        raise HTTPException(status_code=404, detail="Session not found")

    updates: dict[str, Any] = {}
    if request.title is not None:
        updates["title"] = request.title
    # Always update the 'updated' timestamp
    updates["time"] = TimeCreatedUpdated(created=session.time.created, updated=now_ms())
    session = session.model_copy(update=updates)
    state.sessions[session_id] = session  # Update cache
    id_ = state.pool.manifest.config_file_path
    session_data = opencode_to_session_data(session, agent_name=state.agent.name, pool_id=id_)
    await state.storage.save_session(session_data)
    await state.broadcast_event(SessionUpdatedEvent.create(session))
    return session


@router.delete("/{session_id}")
async def delete_session(session_id: str, state: StateDep) -> bool:
    """Delete a session from both cache and storage."""
    # Check if session exists (in cache or storage)
    session = await get_or_load_session(state, session_id)
    if session is None:
        raise HTTPException(status_code=404, detail="Session not found")

    # Cancel any pending permissions and clean up input provider
    if input_provider := state.input_providers.pop(session_id, None):
        input_provider.cancel_all_pending()

    # Remove from cache
    state.sessions.pop(session_id, None)
    state.messages.pop(session_id, None)
    state.session_status.pop(session_id, None)
    state.todos.pop(session_id, None)
    # Delete from storage
    await state.storage.delete_session(session_id)
    await state.broadcast_event(SessionDeletedEvent.create(session_id))
    return True


@router.get("/{session_id}/children")
async def get_session_children(session_id: str, state: StateDep) -> list[Session]:
    """Get all child sessions that were forked from the specified parent session."""
    session = await get_or_load_session(state, session_id)
    if session is None:
        raise HTTPException(status_code=404, detail="Session not found")
    # Search all cached sessions for children
    return [sess for sess in state.sessions.values() if sess.parent_id == session_id]


@router.post("/{session_id}/abort")
async def abort_session(session_id: str, state: StateDep) -> bool:
    """Abort a running session by interrupting the agent."""
    session = await get_or_load_session(state, session_id)
    if session is None:
        raise HTTPException(status_code=404, detail="Session not found")

    # Interrupt the agent to cancel any ongoing stream
    try:
        await state.agent.interrupt()
        # Give a moment for the cancellation to propagate
        await asyncio.sleep(0.1)
    except Exception:  # noqa: BLE001
        pass

    # Update and broadcast session status to notify clients
    state.session_status[session_id] = SessionStatus(type="idle")
    await state.broadcast_event(SessionStatusEvent.create(session_id, SessionStatus(type="idle")))
    return True


@router.post("/{session_id}/fork")
async def fork_session(  # noqa: D417
    session_id: str,
    state: StateDep,
    request: SessionForkRequest | None = None,
    directory: str | None = None,
) -> Session:
    """Fork a session, optionally at a specific message.

    Creates a new session with:
    - parent_id pointing to the original session
    - Copies all messages (or up to message_id if specified)
    - Independent conversation history from that point forward

    Args:
        session_id: The session to fork from
        request: Optional fork parameters (message_id to fork from)
        directory: Optional directory for the forked session

    Returns:
        The newly created forked session
    """
    # Get the original session
    original_session = await get_or_load_session(state, session_id)
    if original_session is None:
        raise HTTPException(status_code=404, detail="Session not found")

    # Get messages from the original session
    original_messages = state.messages.get(session_id, [])
    # Filter messages if message_id is specified
    messages_to_copy: list[MessageWithParts] = []
    if request and request.message_id:
        # Copy messages up to and including the specified message_id
        for msg in original_messages:
            messages_to_copy.append(msg)
            if msg.info.id == request.message_id:
                break
        else:
            # message_id not found in messages
            detail = f"Message {request.message_id} not found in session"
            raise HTTPException(status_code=404, detail=detail)
    else:
        # Copy all messages
        messages_to_copy = list(original_messages)

    # Create the new forked session
    now = now_ms()
    new_session_id = identifier.ascending("session")
    # Use provided directory or inherit from original session
    fork_directory = directory if directory else original_session.directory
    forked_session = Session(
        id=new_session_id,
        project_id=original_session.project_id,
        directory=fork_directory,
        title=f"{original_session.title} (fork)",
        version="1",
        time=TimeCreatedUpdated(created=now, updated=now),
        parent_id=session_id,  # Link to original session
    )

    # Persist the forked session to storage
    session_data = opencode_to_session_data(
        forked_session,
        agent_name=state.agent.name,
        pool_id=state.pool.manifest.config_file_path,
    )
    await state.storage.save_session(session_data)
    # Cache in memory
    state.sessions[new_session_id] = forked_session
    state.session_status[new_session_id] = SessionStatus(type="idle")
    state.todos[new_session_id] = []
    # Copy messages to the new session (with updated session_id references)
    copied_messages: list[MessageWithParts] = []
    for msg_with_parts in messages_to_copy:
        # Create new message info with updated session_id
        new_info = msg_with_parts.info.model_copy(update={"session_id": new_session_id})
        # Copy parts with updated session_id
        new_parts = [
            part.model_copy(update={"session_id": new_session_id}) for part in msg_with_parts.parts
        ]
        copied_messages.append(MessageWithParts(info=new_info, parts=new_parts))

    state.messages[new_session_id] = copied_messages
    input_provider = OpenCodeInputProvider(state, new_session_id)
    state.input_providers[new_session_id] = input_provider
    # Broadcast session created event
    await state.broadcast_event(SessionCreatedEvent.create(forked_session))
    return forked_session


@router.post("/{session_id}/init")
async def init_session(  # noqa: D417
    session_id: str,
    state: StateDep,
    request: SessionInitRequest | None = None,
) -> bool:
    """Initialize a session by analyzing the codebase and creating AGENTS.md.

    Generates a repository map, reads README if present, and runs the agent
    with a prompt to create an AGENTS.md file with project-specific context.

    Args:
        session_id: The session to initialize
        request: Optional model/provider override for the init task

    Returns:
        True when the init task has been started (runs async)
    """
    session = await get_or_load_session(state, session_id)
    if session is None:
        raise HTTPException(status_code=404, detail="Session not found")

    # Get the agent and filesystem
    agent = state.agent
    fs = state.fs
    working_dir = state.working_dir
    try:
        all_files = await find_src_files(fs, working_dir)
        repo_map = RepoMap(fs=fs, root_path=working_dir, max_tokens=4000)
        repomap_content = await repo_map.get_map(all_files) or "No repository map generated."
    except Exception as e:  # noqa: BLE001
        repomap_content = f"Error generating repository map: {e}"

    # Try to read README.md
    readme_content = ""
    for readme_name in ["README.md", "readme.md", "README", "readme.txt"]:
        try:
            readme_path = f"{working_dir}/{readme_name}".replace("//", "/")
            content = await fs._cat_file(readme_path)
            readme_content = content.decode("utf-8") if isinstance(content, bytes) else content
            break
        except Exception:  # noqa: BLE001
            continue

    # Build the init prompt
    prompt_parts = [
        "Please analyze this codebase and create an AGENTS.md file in the project root.",
        "",
        "<repository-structure>",
        repomap_content,
        "</repository-structure>",
    ]
    if readme_content:
        prompt_parts.extend(["", "<readme>", readme_content, "</readme>"])
    prompt_parts.extend([
        "",
        "Include:",
        "1. Build/lint/test commands - especially for running a single test",
        "2. Code style guidelines (imports, formatting, types, naming conventions, error handling)",
        "",
        "The file will be given to AI coding agents working in this repository. "
        "Keep it around 150 lines.",
        "",
        "If there are existing rules (.cursor/rules/, .cursorrules, "
        ".github/copilot-instructions.md), incorporate them.",
    ])

    init_prompt = "\n".join(prompt_parts)

    # Handle model selection if requested
    original_model: str | None = None
    if request and request.model_id and request.provider_id:
        requested_model = f"{request.provider_id}:{request.model_id}"
        try:
            available_models = await agent.get_available_models()
            if available_models:
                valid_ids = [m.id_override if m.id_override else m.id for m in available_models]
                if requested_model in valid_ids:
                    # Store original model to restore later
                    original_model = agent.model_name
                    await agent.set_model(requested_model)
        except Exception:  # noqa: BLE001
            # Agent doesn't support model selection, ignore
            pass

    # Run the agent in the background
    async def run_init() -> None:
        try:
            await agent.run(init_prompt)
        finally:
            # Restore original model if we changed it
            if original_model is not None:
                with contextlib.suppress(Exception):
                    await agent.set_model(original_model)

    state.create_background_task(run_init(), name=f"init_{session_id}")

    return True


@router.get("/{session_id}/todo")
async def get_session_todos(session_id: str, state: StateDep) -> list[Todo]:
    """Get todos for a session.

    Returns todos from the agent pool's TodoTracker.
    """
    session = await get_or_load_session(state, session_id)
    if session is None:
        raise HTTPException(status_code=404, detail="Session not found")

    # Get todos from pool's TodoTracker
    tracker = state.pool.todos
    return [Todo(id=e.id, content=e.content, status=e.status) for e in tracker.entries]


@router.get("/{session_id}/diff")
async def get_session_diff(
    session_id: str,
    state: StateDep,
    message_id: str | None = None,
) -> list[FileDiff]:
    """Get file diffs for a session.

    Returns a list of file changes with unified diffs.
    Optionally filter to changes since a specific message.
    """
    session = await get_or_load_session(state, session_id)
    if session is None:
        raise HTTPException(status_code=404, detail="Session not found")

    file_ops = state.pool.file_ops
    if not file_ops.changes:
        return []
    # Optionally filter by message_id
    changes = file_ops.get_changes_since(message_id) if message_id else file_ops.changes
    return [FileDiff.from_file_change(change) for change in changes]


@router.post("/{session_id}/shell")
async def run_shell_command(
    session_id: str,
    request: ShellRequest,
    state: StateDep,
) -> MessageWithParts:
    """Run a shell command directly."""
    session = await get_or_load_session(state, session_id)
    if session is None:
        raise HTTPException(status_code=404, detail="Session not found")

    # Validate command for security issues
    validate_command(request.command, state.working_dir)
    now = now_ms()
    # Create assistant message for the shell output
    assistant_msg_id = identifier.ascending("message")
    assistant_message = AssistantMessage(
        id=assistant_msg_id,
        session_id=session_id,
        parent_id="",  # Shell commands don't have a parent user message
        model_id=request.model.model_id if request.model else "shell",
        provider_id=request.model.provider_id if request.model else "local",
        mode="shell",
        agent=request.agent,
        path=MessagePath(cwd=state.working_dir, root=state.working_dir),
        time=MessageTime(created=now),
    )

    # Initialize message with empty parts
    assistant_msg_with_parts = MessageWithParts(info=assistant_message, parts=[])
    state.messages[session_id].append(assistant_msg_with_parts)
    # Broadcast message created
    await state.broadcast_event(MessageUpdatedEvent.create(assistant_message))
    # Mark session as busy
    state.session_status[session_id] = SessionStatus(type="busy")
    await state.broadcast_event(SessionStatusEvent.create(session_id, SessionStatus(type="busy")))
    # Add step-start part
    part_id = identifier.ascending("part")
    step_start = StepStartPart(id=part_id, message_id=assistant_msg_id, session_id=session_id)
    assistant_msg_with_parts.parts.append(step_start)
    await state.broadcast_event(PartUpdatedEvent.create(step_start))
    # Execute the command
    output_text = ""
    success = False
    try:
        result = await state.agent.env.execute_command(request.command)
        success = result.success
        if success:
            output_text = str(result.result) if result.result else ""
        else:
            output_text = f"Error: {result.error}" if result.error else "Command failed"
    except Exception as e:  # noqa: BLE001
        output_text = f"Error executing command: {e}"

    response_time = now_ms()
    # Create text part with output
    text_part = TextPart(
        id=identifier.ascending("part"),
        message_id=assistant_msg_id,
        session_id=session_id,
        text=f"$ {request.command}\n{output_text}",
    )
    assistant_msg_with_parts.parts.append(text_part)
    await state.broadcast_event(PartUpdatedEvent.create(text_part))
    part_id = identifier.ascending("part")
    step_finish = StepFinishPart(id=part_id, message_id=assistant_msg_id, session_id=session_id)
    assistant_msg_with_parts.parts.append(step_finish)
    await state.broadcast_event(PartUpdatedEvent.create(step_finish))
    # Update message with completion time
    time_ = MessageTime(created=now, completed=response_time)
    updated_assistant = assistant_message.model_copy(update={"time": time_})
    assistant_msg_with_parts.info = updated_assistant
    await state.broadcast_event(MessageUpdatedEvent.create(updated_assistant))
    # Mark session as idle
    state.session_status[session_id] = SessionStatus(type="idle")
    await state.broadcast_event(SessionStatusEvent.create(session_id, SessionStatus(type="idle")))
    return assistant_msg_with_parts


@router.get("/{session_id}/permissions")
async def get_pending_permissions(
    session_id: str, state: StateDep
) -> list[PermissionAskedProperties]:
    """Get all pending permission requests for a session.

    Returns a list of pending permissions awaiting user response.
    """
    session = await get_or_load_session(state, session_id)
    if session is None:
        raise HTTPException(status_code=404, detail="Session not found")

    # Get the input provider for this session
    input_provider = state.input_providers.get(session_id)
    if input_provider is None:
        return []

    return input_provider.get_pending_permissions()


@router.post("/{session_id}/permissions/{permission_id}")
async def respond_to_permission(
    session_id: str,
    permission_id: str,
    body: PermissionReplyRequest,
    state: StateDep,
) -> bool:
    """Respond to a pending permission request.

    The response can be:
    - "once": Allow this tool execution once
    - "always": Always allow this tool (remembered for session)
    - "reject": Reject this tool execution
    """
    session = await get_or_load_session(state, session_id)
    if session is None:
        raise HTTPException(status_code=404, detail="Session not found")

    # Get the input provider for this session
    input_provider = state.input_providers.get(session_id)
    if input_provider is None:
        raise HTTPException(status_code=404, detail="No input provider for session")

    # Resolve the permission
    resolved = input_provider.resolve_permission(permission_id, body.reply)
    if not resolved:
        raise HTTPException(status_code=404, detail="Permission not found or already resolved")
    event = PermissionResolvedEvent.create(
        session_id=session_id,
        request_id=permission_id,
        reply=body.reply,
    )
    await state.broadcast_event(event)
    return True


# OpenCode-style continuation prompt for summarization
SUMMARIZE_PROMPT = """Provide a detailed prompt for continuing our conversation above. Focus on information that would be helpful for continuing the conversation, including what we did, what we're doing, which files we're working on, and what we're going to do next considering new session will not have access to our conversation."""  # noqa: E501


@router.post("/{session_id}/summarize")
async def summarize_session(  # noqa: PLR0915
    session_id: str,
    state: StateDep,
    request: SummarizeRequest | None = None,
) -> MessageWithParts:
    """Summarize the session conversation.

    First runs the compaction pipeline to condense older messages,
    then streams an LLM-generated summary/continuation prompt to the user.
    The summary message is marked with summary=true for UI display.
    """
    from pydantic_ai.messages import (
        PartDeltaEvent,
        PartStartEvent,
        TextPart as PydanticTextPart,
        TextPartDelta,
    )

    from agentpool.agents.events import StreamCompleteEvent
    from agentpool.messaging.compaction import compact_conversation, summarizing_context

    session = await get_or_load_session(state, session_id)
    if session is None:
        raise HTTPException(status_code=404, detail="Session not found")
    if not state.messages.get(session_id):
        raise HTTPException(status_code=400, detail="No messages to summarize")

    # Determine model to use
    model_id = request.model_id if request and request.model_id else "default"
    provider_id = request.provider_id if request and request.provider_id else "agentpool"

    now = now_ms()
    # Create assistant message for the summary (marked with summary=true)
    assistant_msg_id = identifier.ascending("message")
    assistant_message = AssistantMessage(
        id=assistant_msg_id,
        session_id=session_id,
        parent_id="",
        model_id=model_id,
        provider_id=provider_id,
        mode="summarize",
        agent="summarizer",
        path=MessagePath(cwd=state.working_dir, root=state.working_dir),
        time=MessageTime(created=now),
        summary=True,  # Mark as summary message
    )

    assistant_msg_with_parts = MessageWithParts(info=assistant_message, parts=[])
    state.messages[session_id].append(assistant_msg_with_parts)
    # Broadcast message created
    await state.broadcast_event(MessageUpdatedEvent.create(assistant_message))
    # Mark session as busy
    state.session_status[session_id] = SessionStatus(type="busy")
    await state.broadcast_event(SessionStatusEvent.create(session_id, SessionStatus(type="busy")))
    # Add step-start part
    part_id = identifier.ascending("part")
    step_start = StepStartPart(id=part_id, message_id=assistant_msg_id, session_id=session_id)
    assistant_msg_with_parts.parts.append(step_start)
    await state.broadcast_event(PartUpdatedEvent.create(step_start))
    # Step 1: Stream LLM summary generation FIRST (while we have full history)
    # The LLM sees the complete conversation and generates a continuation prompt.
    response_text = ""
    usage = None
    cost = 0.0
    text_part: TextPart | None = None
    try:
        # Stream events from the agent with the summarization prompt
        # This runs with FULL history - the summary is based on complete context
        async for event in state.agent.run_stream(SUMMARIZE_PROMPT):
            match event:
                # Text streaming start
                case PartStartEvent(part=PydanticTextPart(content=delta)):
                    response_text = delta
                    text_part = TextPart(
                        id=identifier.ascending("part"),
                        message_id=assistant_msg_id,
                        session_id=session_id,
                        text=delta,
                    )
                    assistant_msg_with_parts.parts.append(text_part)
                    await state.broadcast_event(PartUpdatedEvent.create(text_part, delta=delta))

                # Text streaming delta
                case PartDeltaEvent(delta=TextPartDelta(content_delta=delta)) if delta:
                    response_text += delta
                    if text_part is not None:
                        text_part = TextPart(
                            id=text_part.id,
                            message_id=assistant_msg_id,
                            session_id=session_id,
                            text=response_text,
                        )
                        # Update in parts list
                        for i, p in enumerate(assistant_msg_with_parts.parts):
                            if isinstance(p, TextPart) and p.id == text_part.id:
                                assistant_msg_with_parts.parts[i] = text_part
                                break
                        await state.broadcast_event(PartUpdatedEvent.create(text_part, delta=delta))

                # Stream complete - extract token usage
                case StreamCompleteEvent(message=msg) if msg and msg.usage:
                    usage = msg.usage
                    cost = float(msg.cost_info.total_cost) if msg.cost_info else 0

    except Exception as e:  # noqa: BLE001
        response_text = f"Error generating summary: {e}"

    response_time = now_ms()
    # Create/update text part with final response
    if text_part is None:
        text_part = TextPart(
            id=identifier.ascending("part"),
            message_id=assistant_msg_id,
            session_id=session_id,
            text=response_text,
        )
        assistant_msg_with_parts.parts.append(text_part)
        await state.broadcast_event(PartUpdatedEvent.create(text_part))

    # Step 2: Run compaction pipeline AFTER summary is generated
    # The summary was generated with full context. Now we compact the history.
    # Final state will be: [compacted history] + [summary message]
    # The compacted history becomes the cached prefix for future LLM calls.
    try:
        # Get the compaction pipeline from the agent pool configuration
        pipeline = None
        if state.agent.agent_pool is not None:
            pipeline = state.agent.agent_pool.compaction_pipeline
        if pipeline is None:
            # Fall back to a default summarizing pipeline
            pipeline = summarizing_context()

        # Apply the compaction pipeline (modifies agent.conversation in place)
        await compact_conversation(pipeline, state.agent.conversation)
        # Persist compacted messages to storage, replacing the old ones
        if state.storage is not None:
            compacted_history = state.agent.conversation.get_history()
            await state.storage.replace_conversation_messages(session_id, compacted_history)
        # Update in-memory OpenCode messages list with compacted versions
        # Keep only the summary message we just created
        state.messages[session_id] = [assistant_msg_with_parts]

    except Exception:  # noqa: BLE001
        # Compaction failure is not fatal - we still have the summary
        pass
    tokens = Tokens.from_pydantic_ai(usage) if usage else Tokens()
    # Add step-finish part
    step_finish = StepFinishPart(
        id=identifier.ascending("part"),
        message_id=assistant_msg_id,
        session_id=session_id,
        tokens=tokens,
        cost=cost,
    )
    assistant_msg_with_parts.parts.append(step_finish)
    await state.broadcast_event(PartUpdatedEvent.create(step_finish))
    # Update message with completion time and tokens
    msg_time = MessageTime(created=now, completed=response_time)
    update = {"time": msg_time, "tokens": tokens, "cost": cost}
    updated_assistant = assistant_message.model_copy(update=update)
    assistant_msg_with_parts.info = updated_assistant
    await state.broadcast_event(MessageUpdatedEvent.create(updated_assistant))
    # Mark session as idle
    state.session_status[session_id] = SessionStatus(type="idle")
    await state.broadcast_event(SessionStatusEvent.create(session_id, SessionStatus(type="idle")))

    # Broadcast session.diff event after summarization
    file_ops = state.pool.file_ops
    diffs = [FileDiff.from_file_change(change) for change in file_ops.changes]
    await state.broadcast_event(SessionDiffEvent.create(session_id, diffs))

    return assistant_msg_with_parts


@router.post("/{session_id}/share")
async def share_session(
    session_id: str,
    state: StateDep,
    num_messages: int | None = None,
) -> Session:
    """Share session conversation history via OpenCode's sharing service.

    Uses the OpenCode share API to create a shareable link with the full
    conversation including messages and parts.

    Returns the updated session with the share URL.
    """
    session = await get_or_load_session(state, session_id)
    if session is None:
        raise HTTPException(status_code=404, detail="Session not found")
    messages = state.messages.get(session_id, [])

    if not messages:
        raise HTTPException(status_code=400, detail="No messages to share")

    # Apply message limit if specified
    if num_messages is not None and num_messages > 0:
        messages = messages[-num_messages:]
    # Convert our messages to OpenCode Message format
    opencode_messages: list[Message] = []
    for msg_with_parts in messages:
        # Extract text parts
        parts = [
            MessagePart(type="text", text=part.text)
            for part in msg_with_parts.parts
            if isinstance(part, TextPart) and part.text
        ]
        if parts:
            opencode_messages.append(Message(role=msg_with_parts.info.role, parts=parts))
    if not opencode_messages:
        raise HTTPException(status_code=400, detail="No content to share")

    # Share via OpenCode API
    async with OpenCodeSharer() as sharer:
        result = await sharer.share_conversation(opencode_messages, title=session.title)
        share_url = result.url
    # Store the share URL in the session
    share_info = SessionShare(url=share_url)
    updated_session = session.model_copy(update={"share": share_info})
    state.sessions[session_id] = updated_session
    # Broadcast session update
    await state.broadcast_event(SessionUpdatedEvent.create(updated_session))
    return updated_session


class RevertRequest(OpenCodeBaseModel):
    """Request body for reverting a message."""

    message_id: str
    part_id: str | None = None


@router.post("/{session_id}/revert")
async def revert_session(session_id: str, request: RevertRequest, state: StateDep) -> Session:
    """Revert file changes and messages from a specific message.

    Removes messages from the revert point onwards and restores files to their
    state before the specified message's changes.
    """
    from agentpool_server.opencode_server.models import MessageRemovedEvent, PartRemovedEvent

    session = await get_or_load_session(state, session_id)
    if session is None:
        raise HTTPException(status_code=404, detail="Session not found")

    # Get messages for this session
    messages = state.messages.get(session_id, [])
    if not messages:
        raise HTTPException(status_code=400, detail="No messages to revert")

    # Find the revert message index
    revert_index = None
    for i, msg in enumerate(messages):
        if msg.info.id == request.message_id:
            revert_index = i
            break

    if revert_index is None:
        raise HTTPException(status_code=404, detail=f"Message {request.message_id} not found")

    # Split messages: keep messages before revert point, remove from revert point onwards
    messages_to_keep = messages[:revert_index]
    messages_to_remove = messages[revert_index:]

    if not messages_to_remove:
        raise HTTPException(status_code=400, detail="No messages to revert")

    # Store removed messages for unrevert
    state.reverted_messages[session_id] = messages_to_remove
    # Update message list - keep only messages before revert point
    state.messages[session_id] = messages_to_keep
    # Emit message.removed and part.removed events for all removed messages
    for msg in messages_to_remove:
        # Emit message.removed event
        await state.broadcast_event(MessageRemovedEvent.create(session_id, msg.info.id))

        # Emit part.removed events for all parts
        for part in msg.parts:
            await state.broadcast_event(PartRemovedEvent.create(session_id, msg.info.id, part.id))

    # Also revert file changes if any
    file_ops = state.pool.file_ops
    if file_ops.changes and (
        revert_ops := file_ops.get_revert_operations(since_message_id=request.message_id)
    ):
        for path, content in revert_ops:
            try:
                if content is None:
                    await state.fs._rm_file(path)
                else:
                    content_bytes = content.encode("utf-8")
                    await state.fs._pipe_file(path, content_bytes)
            except Exception as e:
                detail = f"Failed to revert {path}: {e}"
                raise HTTPException(status_code=500, detail=detail) from e
        file_ops.remove_changes_since(request.message_id)

    # Update session with revert info
    session = state.sessions[session_id]
    revert_info = SessionRevert(message_id=request.message_id, part_id=request.part_id)
    updated_session = session.model_copy(update={"revert": revert_info})
    state.sessions[session_id] = updated_session

    # Broadcast session update
    await state.broadcast_event(SessionUpdatedEvent.create(updated_session))

    # Broadcast session.diff event with current file diffs
    file_ops = state.pool.file_ops
    diffs = [FileDiff.from_file_change(change) for change in file_ops.changes]
    await state.broadcast_event(SessionDiffEvent.create(session_id, diffs))

    return updated_session


@router.post("/{session_id}/unrevert")
async def unrevert_session(session_id: str, state: StateDep) -> Session:
    """Restore all reverted messages and file changes.

    Re-applies the messages and changes that were previously reverted.
    """
    from agentpool_server.opencode_server.models import MessageUpdatedEvent, PartUpdatedEvent

    session = await get_or_load_session(state, session_id)
    if session is None:
        raise HTTPException(status_code=404, detail="Session not found")

    # Restore reverted messages
    reverted_messages = state.reverted_messages.get(session_id, [])
    if not reverted_messages:
        raise HTTPException(status_code=400, detail="No reverted messages to restore")

    # Restore messages to conversation
    if session_id not in state.messages:
        state.messages[session_id] = []
    state.messages[session_id].extend(reverted_messages)

    # Emit message.updated and part.updated events for restored messages
    for msg in reverted_messages:
        # Emit message.updated event
        await state.broadcast_event(MessageUpdatedEvent.create(msg.info))
        # Emit part.updated events for all parts
        for part in msg.parts:
            await state.broadcast_event(PartUpdatedEvent.create(part))

    # Clear reverted messages
    state.reverted_messages.pop(session_id, None)

    # Also unrevert file changes if any
    file_ops = state.pool.file_ops
    if file_ops.reverted_changes:
        unrevert_ops = file_ops.get_unrevert_operations()
        for path, content in unrevert_ops:
            try:
                if content is None:
                    await state.fs._rm_file(path)
                else:
                    content_bytes = content.encode("utf-8")
                    await state.fs._pipe_file(path, content_bytes)
            except Exception as e:
                detail = f"Failed to unrevert {path}: {e}"
                raise HTTPException(status_code=500, detail=detail) from e
        file_ops.restore_reverted_changes()

    # Clear revert info from session
    updated_session = session.model_copy(update={"revert": None})
    state.sessions[session_id] = updated_session
    # Broadcast session update
    await state.broadcast_event(SessionUpdatedEvent.create(updated_session))
    return updated_session


@router.delete("/{session_id}/share")
async def unshare_session(session_id: str, state: StateDep) -> bool:
    """Remove share link from a session.

    Note: This only removes the link from the session metadata.
    The shared content may still exist on the provider's servers.
    """
    session = await get_or_load_session(state, session_id)
    if session is None:
        raise HTTPException(status_code=404, detail="Session not found")
    if session.share is None:
        raise HTTPException(status_code=400, detail="Session is not shared")
    # Remove share info from session
    updated_session = session.model_copy(update={"share": None})
    state.sessions[session_id] = updated_session
    # Broadcast session update
    await state.broadcast_event(SessionUpdatedEvent.create(updated_session))
    return True


@router.post("/{session_id}/command")
async def execute_command(  # noqa: PLR0915
    session_id: str,
    request: CommandRequest,
    state: StateDep,
) -> MessageWithParts:
    """Execute a slash command (MCP prompt).

    Commands are mapped to MCP prompts. The command name is used to find
    the matching prompt, and arguments are parsed and passed to it.
    """
    session = await get_or_load_session(state, session_id)
    if session is None:
        raise HTTPException(status_code=404, detail="Session not found")
    prompts = await state.agent.tools.list_prompts()
    # Find matching prompt by name
    prompt = next((p for p in prompts if p.name == request.command), None)
    if prompt is None:
        detail = f"Command not found: {request.command}"
        raise HTTPException(status_code=404, detail=detail)

    # Parse arguments - OpenCode uses $1, $2 style, MCP uses named arguments
    # For simplicity, we'll pass the raw arguments string to the first argument
    # or parse space-separated args into a dict
    arguments: dict[str, str] = {}
    if request.arguments and prompt.arguments:
        # Split arguments and map to prompt argument names
        arg_values = request.arguments.split()
        for i, arg_def in enumerate(prompt.arguments):
            if i < len(arg_values):
                arguments[arg_def["name"]] = arg_values[i]

    now = now_ms()
    # Create assistant message
    assistant_msg_id = identifier.ascending("message")
    assistant_message = AssistantMessage(
        id=assistant_msg_id,
        session_id=session_id,
        parent_id="",
        model_id=request.model or "default",
        provider_id="mcp",
        mode="command",
        agent=request.agent or "default",
        path=MessagePath(cwd=state.working_dir, root=state.working_dir),
        time=MessageTime(created=now),
    )
    assistant_msg_with_parts = MessageWithParts(info=assistant_message, parts=[])
    state.messages[session_id].append(assistant_msg_with_parts)
    await state.broadcast_event(MessageUpdatedEvent.create(assistant_message))
    # Mark session as busy
    state.session_status[session_id] = SessionStatus(type="busy")
    await state.broadcast_event(SessionStatusEvent.create(session_id, SessionStatus(type="busy")))
    # Add step-start part
    part_id = identifier.ascending("part")
    step_start = StepStartPart(id=part_id, message_id=assistant_msg_id, session_id=session_id)
    assistant_msg_with_parts.parts.append(step_start)
    await state.broadcast_event(PartUpdatedEvent.create(step_start))

    # Get prompt content and execute through the agent
    try:
        prompt_parts = await prompt.get_components(arguments)
        # Extract text content from parts
        prompt_texts = []
        for part in prompt_parts:
            if hasattr(part, "content"):
                content = part.content
                if isinstance(content, str):
                    prompt_texts.append(content)
                elif isinstance(content, list):
                    # Handle Sequence[UserContent]
                    for item in content:
                        if isinstance(item, FileUrl):
                            prompt_texts.append(item.url)
                        elif isinstance(item, str):
                            prompt_texts.append(item)
        prompt_text = "\n".join(prompt_texts)
        # Run the expanded prompt through the agent
        result = await state.agent.run(prompt_text)
        output_text = str(result.data)

    except Exception as e:  # noqa: BLE001
        output_text = f"Error executing command: {e}"

    response_time = now_ms()
    # Create text part with output
    text_part = TextPart(
        id=identifier.ascending("part"),
        message_id=assistant_msg_id,
        session_id=session_id,
        text=output_text,
    )
    assistant_msg_with_parts.parts.append(text_part)
    await state.broadcast_event(PartUpdatedEvent.create(text_part))
    step_finish = StepFinishPart(
        id=identifier.ascending("part"),
        message_id=assistant_msg_id,
        session_id=session_id,
    )
    assistant_msg_with_parts.parts.append(step_finish)
    await state.broadcast_event(PartUpdatedEvent.create(step_finish))
    # Update message with completion time
    time_ = MessageTime(created=now, completed=response_time)
    updated_assistant = assistant_message.model_copy(update={"time": time_})
    assistant_msg_with_parts.info = updated_assistant
    await state.broadcast_event(MessageUpdatedEvent.create(updated_assistant))
    # Mark session as idle
    state.session_status[session_id] = SessionStatus(type="idle")
    await state.broadcast_event(SessionStatusEvent.create(session_id, SessionStatus(type="idle")))

    # Broadcast command.executed event
    await state.broadcast_event(
        CommandExecutedEvent.create(
            name=request.command,
            session_id=session_id,
            arguments=request.arguments or "",
            message_id=assistant_msg_id,
        )
    )

    return assistant_msg_with_parts



"""Fetch agent data from the ACP registry."""

from __future__ import annotations

from typing import Final

import httpx

from acp.registry.model import DistributionUnion, Registry, RegistryAgent, UvxDistribution


REGISTRY_URL: Final[str] = "https://cdn.agentclientprotocol.com/registry/v1/latest/registry.json"

_BUILTIN_AGENTS: Final[tuple[RegistryAgent, ...]] = (
    RegistryAgent(
        id="dummy",
        name="Dummy",
        version="0.2.0",
        description="Built-in ACP dummy agent for protocol testing",
        repository="https://github.com/observerw/acp-agent",
        authors=["observerw"],
        license="MIT",
        distribution=DistributionUnion(
            uvx=UvxDistribution(package="acp-agent", args=["dummy"]),
        ),
    ),
)


def _builtin_agent(agent_id: str) -> RegistryAgent | None:
    return next((agent for agent in _BUILTIN_AGENTS if agent.id == agent_id), None)


def _merge_builtin_agents(agents: list[RegistryAgent]) -> list[RegistryAgent]:
    existing_ids = {agent.id for agent in agents}
    extras = [agent for agent in _BUILTIN_AGENTS if agent.id not in existing_ids]
    return [*agents, *extras]


async def fetch_registry() -> Registry:
    """Fetch the ACP registry data."""
    async with httpx.AsyncClient() as client:
        response = await client.get(REGISTRY_URL, headers={"User-Agent": "Mozilla/5.0"})
        response.raise_for_status()
        data = response.json()
    return Registry.model_validate(data)


async def list_agents() -> list[RegistryAgent]:
    """List all agents from the ACP registry, including built-in agents."""
    registry = await fetch_registry()
    return _merge_builtin_agents(registry.agents)


async def fetch_agent(agent_id: str) -> RegistryAgent | None:
    """Fetch a single agent by ID from the registry or built-in agents."""
    if agent := _builtin_agent(agent_id):
        return agent
    agents = await list_agents()
    return next((a for a in agents if a.id == agent_id), None)


if __name__ == "__main__":

    async def main() -> None:
        agents = await list_agents()
        print(agents)

    import asyncio

    asyncio.run(main())



"""OpenCode-compatible FastAPI server.

This server implements the OpenCode API endpoints to allow OpenCode SDK clients
to interact with AgentPool agents.
"""

from __future__ import annotations

from contextlib import asynccontextmanager
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any

from fastapi import FastAPI, Request  # noqa: TC002
from fastapi.exceptions import RequestValidationError
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse, RedirectResponse, Response

from agentpool import log
from agentpool_server.opencode_server.routes import (
    agent_router,
    app_router,
    config_router,
    file_router,
    global_router,
    lsp_router,
    message_router,
    permission_router,
    pty_router,
    question_router,
    session_router,
    tui_router,
)
from agentpool_server.opencode_server.state import ServerState


if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Set as AbstractSet

    from httpx import Headers

    from agentpool.agents.base_agent import BaseAgent
    from agentpool.storage.manager import SessionMetadataGeneratedEvent
    from agentpool.utils.todos import TodoTracker


VERSION = "0.1.0"
logger = log.get_logger(__name__)


def filter_headers(headers: Headers) -> dict[str, str]:
    excluded_headers = {"content-encoding", "content-length", "transfer-encoding", "connection"}
    return {k: v for k, v in headers.items() if k.lower() not in excluded_headers}


class OpenCodeJSONResponse(JSONResponse):
    """Custom JSON response that excludes None values (like OpenCode does)."""

    def render(self, content: Any) -> bytes:
        from fastapi.encoders import jsonable_encoder

        return super().render(jsonable_encoder(content, exclude_none=True))


async def check_pypi_version(package: str = "agentpool") -> str | None:
    """Check PyPI for the latest version of a package.

    Args:
        package: Package name to check

    Returns:
        Latest version string, or None if check fails
    """
    import httpx

    try:
        async with httpx.AsyncClient(timeout=5.0) as client:
            response = await client.get(f"https://pypi.org/pypi/{package}/json")
            if response.status_code == 200:  # noqa: PLR2004
                data: dict[str, Any] = response.json()
                info: dict[str, Any] = data.get("info", {})
                version: str | None = info.get("version")
                return version
    except Exception:  # noqa: BLE001
        pass
    return None


def compare_versions(current: str, latest: str) -> bool:
    """Check if latest version is newer than current."""
    from packaging.version import Version

    try:
        return Version(latest) > Version(current)
    except Exception:  # noqa: BLE001
        return False


def create_app(*, agent: BaseAgent[Any, Any], working_dir: str | None = None) -> FastAPI:  # noqa: PLR0915
    """Create the FastAPI application.

    Args:
        agent: The agent to use for handling messages. Must have agent_pool set.
        working_dir: Working directory for file operations. Defaults to cwd.

    Returns:
        Configured FastAPI application.

    Raises:
        ValueError: If agent has no agent_pool set.
    """
    import logfire

    if agent.agent_pool is None:
        msg = "Agent must have agent_pool set"
        raise ValueError(msg)

    state = ServerState(working_dir=working_dir or str(Path.cwd()), agent=agent)

    # Set up todo change callback to broadcast events
    async def on_todo_change(tracker: TodoTracker) -> None:
        """Broadcast todo updates to all active sessions."""
        from agentpool_server.opencode_server.models.events import Todo, TodoUpdatedEvent

        # Convert tracker entries to OpenCode Todo models
        todos = [
            Todo(id=e.id, content=e.content, status=e.status, priority=e.priority)
            for e in tracker.entries
        ]
        # Broadcast to all active sessions
        for session_id in state.sessions:
            event = TodoUpdatedEvent.create(session_id=session_id, todos=todos)
            await state.broadcast_event(event)

    state.pool.todos.on_change = on_todo_change
    # Set up title generation callback to update OpenCode sessions

    async def on_title_generated(event: SessionMetadataGeneratedEvent) -> None:
        """Update session when metadata is generated by StorageManager."""
        from agentpool_server.opencode_server.converters import opencode_to_session_data
        from agentpool_server.opencode_server.models.events import SessionUpdatedEvent

        logger.info("on_title_generated called", session_id=event.session_id, data=event.metadata)
        session_id = event.session_id
        if session_id in state.sessions:
            # Update in-memory session
            session = state.sessions[session_id]
            updated_session = session.model_copy(update={"title": event.metadata.title})
            state.sessions[session_id] = updated_session
            # Persist to storage
            session_data = opencode_to_session_data(
                updated_session,
                agent_name=state.agent.name,
                pool_id=state.pool.manifest.config_file_path,
            )
            await state.storage.save_session(session_data)
            # Broadcast session update to UI
            await state.broadcast_event(SessionUpdatedEvent.create(updated_session))
        else:
            logger.warning("Session not found in state.sessions", session_id=session_id)

    # Connect to storage manager's metadata_generated signal
    if state.storage:
        state.storage.metadata_generated.connect(on_title_generated)

    # Watchers for VCS and file events
    branch_watcher: Any = None
    project_file_watcher: Any = None

    @asynccontextmanager
    async def lifespan(app: FastAPI) -> AsyncIterator[None]:  # noqa: PLR0915
        nonlocal branch_watcher, project_file_watcher

        from watchfiles import Change

        from agentpool.utils.file_watcher import FileWatcher, GitBranchWatcher
        from agentpool_server.opencode_server.models.events import (
            FileWatcherUpdatedEvent,
            VcsBranchUpdatedEvent,
        )

        # --- Git branch watcher ---
        async def on_branch_change(branch: str | None) -> None:
            """Broadcast branch change to all subscribers."""
            logger.info("Broadcasting vcs.branch.updated event", branch=branch)
            event = VcsBranchUpdatedEvent.create(branch=branch)
            await state.broadcast_event(event)

        logger.info("Setting up GitBranchWatcher", working_dir=state.working_dir)
        branch_watcher = GitBranchWatcher(repo_path=state.working_dir, callback=on_branch_change)
        await branch_watcher.start()
        logger.info("GitBranchWatcher started", current_branch=branch_watcher.current_branch)
        # --- Project file watcher ---
        # Map watchfiles Change types to OpenCode event types
        change_type_map: dict[Change, str] = {
            Change.added: "add",
            Change.modified: "change",
            Change.deleted: "unlink",
        }

        # Get ignore patterns from config
        ignore_patterns: list[str] = []
        if state.config and state.config.watcher and state.config.watcher.ignore:
            ignore_patterns = state.config.watcher.ignore

        def should_ignore(file_path: str) -> bool:
            """Check if a file path should be ignored."""
            import fnmatch

            # Always ignore .git
            if "/.git/" in file_path or file_path.endswith("/.git"):
                return True
            # Check user-configured patterns
            rel_path = file_path
            if state.working_dir and file_path.startswith(state.working_dir):
                rel_path = file_path[len(state.working_dir) :].lstrip("/")
            return any(fnmatch.fnmatch(rel_path, pat) for pat in ignore_patterns)

        async def on_file_change(changes: AbstractSet[tuple[Change, str]]) -> None:
            """Broadcast file changes to all subscribers."""
            for change_type, file_path in changes:
                if should_ignore(file_path):
                    continue
                event_type = change_type_map.get(change_type, "change")
                logger.info(
                    "Broadcasting file.watcher.updated", event_type=event_type, path=file_path
                )
                event = FileWatcherUpdatedEvent.create(file=file_path, event=event_type)  # type: ignore[arg-type]
                await state.broadcast_event(event)

        logger.info("Setting up project FileWatcher", working_dir=state.working_dir)
        project_file_watcher = FileWatcher(
            paths=[state.working_dir],
            callback=on_file_change,
            debounce=500,  # 500ms debounce to batch rapid changes
        )
        await project_file_watcher.start()
        logger.info("Project FileWatcher started")

        # --- Version update check (triggered when first client connects) ---
        async def check_for_updates() -> None:
            """Check PyPI for updates and notify via toast."""
            from agentpool import __version__ as current_version
            from agentpool_server.opencode_server.models.events import TuiToastShowEvent

            latest = await check_pypi_version("agentpool")
            if latest and compare_versions(current_version, latest):
                logger.info("Update available", current_version=current_version, latest=latest)
                event = TuiToastShowEvent.create(
                    title="Update Available",
                    message=f"agentpool {latest} is available (current: {current_version})",
                    variant="info",
                    duration=10000,
                )
                await state.broadcast_event(event)

        # Register callback to run when first SSE client connects
        state.on_first_subscriber = check_for_updates
        # Pool context is managed externally (by the caller)
        yield
        # Shutdown - clean up
        state.pool.todos.on_change = None
        if branch_watcher:
            await branch_watcher.stop()
        if project_file_watcher:
            await project_file_watcher.stop()
        # Clean up LSP servers
        await state.lsp_manager.stop_all()

    app = FastAPI(
        title="OpenCode-Compatible API",
        description="AgentPool server with OpenCode API compatibility",
        version=VERSION,
        lifespan=lifespan,
        default_response_class=OpenCodeJSONResponse,
    )

    # Add CORS middleware (required for OpenCode TUI)
    app.add_middleware(
        CORSMiddleware,  # ty: ignore[invalid-argument-type]
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    # Store state on app for access in routes
    app.state.server_state = state

    @app.exception_handler(RequestValidationError)
    async def validation_exception_handler(
        request: Request, exc: RequestValidationError
    ) -> JSONResponse:
        body = await request.body()
        print(f"Validation error for {request.url}")
        print(f"Body: {body.decode()}")
        print(f"Errors: {exc.errors()}")
        content = {"detail": exc.errors(), "body": body.decode()}
        return JSONResponse(status_code=422, content=content)

    # Register routers
    app.include_router(global_router)
    app.include_router(app_router)
    app.include_router(config_router)
    app.include_router(session_router)
    app.include_router(message_router)
    app.include_router(file_router)
    app.include_router(agent_router)
    app.include_router(permission_router)
    app.include_router(question_router)
    app.include_router(pty_router)
    app.include_router(tui_router)
    app.include_router(lsp_router)

    # OpenAPI doc redirect
    @app.get("/doc")
    async def get_doc() -> RedirectResponse:
        """Redirect to OpenAPI docs."""
        return RedirectResponse(url="/docs")

    # Proxy catch-all for OpenCode's hosted web UI
    # This must be registered LAST so it doesn't catch API routes
    @app.api_route("/{path:path}", methods=["GET", "HEAD", "OPTIONS"])
    async def proxy_web_ui(request: Request, path: str) -> Response:
        """Proxy unmatched GET requests to OpenCode's hosted web UI.

        This allows users to open http://localhost:4096 in a browser and get
        the full OpenCode web interface, which then makes API calls back to
        this local server for all data operations.
        """
        import httpx

        # Build target URL
        url = f"https://app.opencode.ai/{path}"
        if request.url.query:
            url += f"?{request.url.query}"

        async with httpx.AsyncClient(timeout=30.0) as client:
            # Forward the request
            response = await client.request(
                method=request.method,
                url=url,
                headers={"host": "app.opencode.ai"},
                follow_redirects=True,
            )
            # Filter out hop-by-hop headers that shouldn't be forwarded
            return Response(
                content=response.content,
                status_code=response.status_code,
                headers=filter_headers(response.headers),
                media_type=response.headers.get("content-type"),
            )

    logfire.instrument_fastapi(app)
    return app


@dataclass
class OpenCodeServer:
    """OpenCode-compatible server wrapper.

    Provides a convenient interface for running the server.
    """

    agent: BaseAgent[Any, Any]
    """The agent to use for handling messages. Must have agent_pool set."""
    host: str = "127.0.0.1"
    """Host to bind to."""
    port: int = 4096
    """Port to listen on."""
    working_dir: str | None = None
    """Working directory for file operations."""
    _app: FastAPI | None = field(default=None, init=False, repr=False)

    @property
    def app(self) -> FastAPI:
        """Get or create the FastAPI application."""
        if self._app is None:
            self._app = create_app(agent=self.agent, working_dir=self.working_dir)
        return self._app

    def run(self) -> None:
        """Run the server (blocking)."""
        import uvicorn

        uvicorn.run(self.app, host=self.host, port=self.port)

    async def run_async(self) -> None:
        """Run the server asynchronously."""
        import uvicorn

        config = uvicorn.Config(self.app, host=self.host, port=self.port, ws="websockets-sansio")
        server = uvicorn.Server(config)
        await server.serve()


def run_server(
    agent: BaseAgent[Any, Any],
    *,
    host: str = "127.0.0.1",
    port: int = 4096,
    working_dir: str | None = None,
) -> None:
    """Run the OpenCode-compatible server.

    Args:
        agent: The agent to use for handling messages. Must have agent_pool set.
        host: Host to bind to.
        port: Port to listen on.
        working_dir: Working directory for file operations.
    """
    server = OpenCodeServer(agent, host=host, port=port, working_dir=working_dir)
    server.run()


if __name__ == "__main__":
    import asyncio

    from agentpool import AgentPool, config_resources

    async def main() -> None:
        pool = AgentPool(config_resources.CLAUDE_CODE_ASSISTANT)
        async with pool:
            run_server(pool.main_agent)

    asyncio.run(main())



"""Streaming Tools - Base classes for tools that use the current agent for streaming.

This module provides a pattern for creating tools that:
1. Take a description/intent as input from the outer agent
2. Use the SAME agent with forked history for streaming
3. Process that stream with custom logic (UI updates, incremental writes, etc.)

The key insight is that this sidesteps the limitation of pydantic-ai where tool
execution happens AFTER the model finishes streaming. By using the same agent
with forked history, we get:
- Streaming "for free" while maintaining the tool abstraction
- Full conversation context preserved (the agent knows what was discussed)
- Cache benefits from the shared conversation history

Cache behavior: The forked conversation shares the same cache as the outer agent
since it uses identical message history. The fork's new messages are NOT persisted
back to the main conversation.

Example usage:

    @streaming_tool(
        system_prompt="Output the complete file content.",
    )
    async def edit_file(
        ctx: AgentContext,
        stream: AsyncIterator[StreamChunk],
        path: str,
    ) -> str:
        content = ""
        async for chunk in stream:
            content += chunk.text
            await ctx.events.file_edit_progress(path, content, "in_progress")

        await write_file(path, content)
        return f"Edited {path}"

    # Exposed to outer agent as: edit_file(description: str, path: str) -> str
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
import functools
from typing import TYPE_CHECKING, Any

from pydantic_ai import PartDeltaEvent, TextPartDelta


if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Callable

    from pydantic_ai.messages import ModelRequest, ModelResponse

    from agentpool.agents.base_agent import BaseAgent
    from agentpool.agents.context import AgentContext
    from agentpool.messaging import MessageHistory


@dataclass
class StreamChunk:
    """A chunk of streamed text from the agent."""

    text: str
    """The text content of this chunk."""

    accumulated: str
    """All text accumulated so far."""

    is_complete: bool = False
    """Whether this is the final chunk."""


@dataclass
class StreamingToolConfig:
    """Configuration for a streaming tool."""

    prompt_template: str | None = None
    """Optional prompt template. Use {description} and {kwargs} placeholders."""

    include_file_content: bool = False
    """Whether to include file content in the prompt (for edit tools)."""


class StreamingToolBase(ABC):
    """Abstract base class for streaming tools.

    Subclass this to create tools that use the current agent for streaming.
    """

    config: StreamingToolConfig

    @abstractmethod
    async def process_stream(
        self,
        ctx: AgentContext,
        stream: AsyncIterator[StreamChunk],
        **kwargs: Any,
    ) -> str:
        """Process the streamed output, return final result for outer agent."""
        ...

    @abstractmethod
    def build_prompt(self, description: str, **kwargs: Any) -> str:
        """Build the prompt for the streaming request."""
        ...

    async def execute(self, ctx: AgentContext, description: str, **kwargs: Any) -> str:
        """Execute the streaming tool.

        This is called by the outer agent. It:
        1. Forks the current conversation history
        2. Streams a response using the same agent
        3. Passes the stream to process_stream()
        """
        agent = ctx.native_agent
        fork_history = _create_forked_history(agent)
        prompt = self.build_prompt(description, **kwargs)
        stream = agent.run_stream(prompt, message_history=fork_history, store_history=False)
        chunk_stream = _create_chunk_stream(stream)
        return await self.process_stream(ctx, chunk_stream, **kwargs)


def streaming_tool(
    prompt_template: str | None = None,
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """Decorator that wraps a stream-processing function into a streaming tool.

    The decorated function should have the signature:
        async def fn(ctx: AgentContext, stream: AsyncIterator[StreamChunk], **kwargs) -> str

    The resulting tool will have the signature:
        async def fn(ctx: AgentContext, description: str, **kwargs) -> str

    Args:
        prompt_template: Optional template for building the prompt.
            Use {description} for the description and {key} for kwargs.

    Example:
        @streaming_tool()
        async def generate_code(
            ctx: AgentContext,
            stream: AsyncIterator[StreamChunk],
            language: str,
        ) -> str:
            code = ""
            async for chunk in stream:
                code = chunk.accumulated
                await ctx.events.progress(code)
            return f"Generated {len(code)} chars of {language} code"
    """

    def decorator(fn: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(fn)
        async def wrapper(ctx: AgentContext, description: str, **kwargs: Any) -> str:
            agent = ctx.native_agent
            fork_history = _create_forked_history(agent)
            # Build prompt
            if prompt_template:
                prompt = prompt_template.format(description=description, **kwargs)
            else:
                prompt = _build_default_prompt(description, **kwargs)
            # Stream using the same agent with forked history
            stream = agent.run_stream(prompt, message_history=fork_history, store_history=False)
            chunk_stream = _create_chunk_stream(stream)
            return await fn(ctx, chunk_stream, **kwargs)  # type: ignore[no-any-return]

        # Mark as streaming tool for introspection
        wrapper._streaming_tool = True  # type: ignore[attr-defined]
        wrapper._prompt_template = prompt_template  # type: ignore[attr-defined]
        return wrapper

    return decorator


def _create_forked_history(agent: BaseAgent) -> MessageHistory:
    """Create a forked message history from the current conversation.

    The fork preserves full context for cache benefits while isolating
    new messages from the main conversation.
    """
    from agentpool.messaging import ChatMessage, MessageHistory

    if current_history := agent.conversation.get_history():
        # Get pydantic-ai messages and inject cache point
        all_messages: list[ModelRequest | ModelResponse] = []
        for msg in current_history:
            all_messages.extend(msg.to_pydantic_ai())
        # Wrap in a single ChatMessage for the forked history
        msg = ChatMessage(messages=all_messages, role="user", content="")
        return MessageHistory(messages=[msg])

    return MessageHistory()


async def _create_chunk_stream(stream: Any) -> AsyncIterator[StreamChunk]:
    """Create an async iterator of StreamChunk from an agent stream."""
    accumulated = ""

    async for node in stream:
        match node:
            case PartDeltaEvent(delta=TextPartDelta(content_delta=text)):
                accumulated += text
                yield StreamChunk(text=text, accumulated=accumulated, is_complete=False)

    yield StreamChunk(text="", accumulated=accumulated, is_complete=True)


def _build_default_prompt(description: str, **kwargs: Any) -> str:
    """Build a default prompt from description and kwargs."""
    parts = [description]
    # Add context from kwargs if relevant
    for key, value in kwargs.items():
        if value is not None:
            parts.append(f"{key}: {value}")

    return "\n\n".join(parts)


# --- Example Usage ---
#
# @streaming_tool()
# async def streaming_write_file(
#     ctx: AgentContext,
#     stream: AsyncIterator[StreamChunk],
#     path: str,
# ) -> str:
#     """Write a file with streaming progress updates."""
#     content = ""
#
#     async for chunk in stream:
#         content = chunk.accumulated
#         await ctx.events.file_edit_progress(
#             path=path,
#             old_text="",
#             new_text=content,
#             status="streaming" if not chunk.is_complete else "completed",
#         )
#
#     # Write the file
#     from pathlib import Path
#     Path(path).write_text(content)
#     return f"Successfully created {path}"



"""Claude Code storage provider.

This module implements storage compatible with Claude Code's filesystem format,
enabling interoperability between agentpool and Claude Code.

Key features:
- JSONL-based conversation logs per project
- Multi-agent support (main + sub-agents)
- Message ancestry tracking
- Conversation forking and branching

See ARCHITECTURE.md for detailed documentation of the storage format.
"""

from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any

import anyenv
from clawd_code_sdk.storage.helpers import (
    count_session_messages,
    encode_project_path,
    extract_title,
    read_session,
    write_entry,
)
from clawd_code_sdk.storage.models import (
    ClaudeAssistantEntry,
    ClaudeEntry,
    ClaudeToolUseBlock,
    ClaudeUserEntry,
)

from agentpool.log import get_logger
from agentpool.utils.thread_helpers import parallel_map
from agentpool.utils.time_utils import get_now, parse_iso_timestamp
from agentpool_config.storage import ClaudeStorageConfig
from agentpool_storage.base import StorageProvider
from agentpool_storage.claude_provider.converters import (
    chat_message_to_entry,
    entry_to_chat_message,
    normalize_model_name,
)
from agentpool_storage.models import TokenUsage


if TYPE_CHECKING:
    from clawd_code_sdk.storage.models import ClaudeJSONLEntry

    from agentpool.messaging import ChatMessage
    from agentpool.sessions.models import SessionData
    from agentpool_config.session import SessionQuery
    from agentpool_storage.models import ConversationData, QueryFilters, StatsFilters


logger = get_logger(__name__)


def _build_tool_id_mapping(entries: list[ClaudeJSONLEntry]) -> dict[str, str]:
    """Build a mapping from tool_call_id to tool_name from assistant entries."""
    mapping: dict[str, str] = {}
    for entry in entries:
        if not isinstance(entry, ClaudeAssistantEntry):
            continue
        content = entry.message.content
        if not isinstance(content, list):
            continue
        for block in content:
            if isinstance(block, ClaudeToolUseBlock) and block.id and block.name:
                mapping[block.id] = block.name
    return mapping


@dataclass(slots=True)
class SessionMetadata:
    """Lightweight session metadata without full message parsing."""

    session_id: str
    path: Path
    first_timestamp: str | None
    last_timestamp: str | None
    cwd: str | None
    message_count: int
    title: str | None


@dataclass(slots=True)
class ParsedSession:
    """Fully parsed session with entries, messages, and metadata."""

    session_id: str
    path: Path
    entries: list[ClaudeJSONLEntry]
    tool_mapping: dict[str, str]
    messages: list[ChatMessage[str]]
    first_timestamp: datetime | None
    total_tokens: int


def _read_session_metadata(
    session_path: Path,
    max_title_chars: int = 60,
) -> SessionMetadata | None:
    """Read minimal metadata from a session file without full parsing.

    Extracts timestamps, message count, cwd, and title in a single pass.
    This is much faster than validating all entries with Pydantic.

    Args:
        session_path: Path to the JSONL session file
        max_title_chars: Maximum characters for title (truncates with '...')

    Returns:
        SessionMetadata with timestamps, message count, and title,
        or None if file is empty/invalid
    """
    if not session_path.exists():
        return None

    first_timestamp: str | None = None
    last_timestamp: str | None = None
    cwd: str | None = None
    title: str | None = None
    message_count = 0

    with session_path.open("r", encoding="utf-8") as f:
        for line in f:
            stripped = line.strip()
            if not stripped:
                continue
            try:
                data = anyenv.load_json(stripped, return_type=dict)
                entry_type = data.get("type")

                # Summary entries provide the best title
                if entry_type == "summary" and title is None and (summary := data.get("summary")):
                    title = str(summary)

                # Count user/assistant messages and extract metadata
                if entry_type in ("user", "assistant"):
                    message_count += 1
                    timestamp = data.get("timestamp")
                    # Get timestamp and cwd from first message entry
                    if first_timestamp is None:
                        first_timestamp = timestamp
                        if cwd is None:
                            cwd = data.get("cwd")

                    # First user message as fallback title
                    if entry_type == "user" and title is None:
                        msg = data.get("message", {})
                        content = msg.get("content", "")
                        if isinstance(content, str) and content:
                            first_line = content.split("\n")[0].strip()
                            if first_line:
                                if len(first_line) > max_title_chars:
                                    title = first_line[:max_title_chars] + "..."
                                else:
                                    title = first_line

                    # Always update last_timestamp for user/assistant messages
                    last_timestamp = timestamp
            except anyenv.JsonLoadError:
                continue

    if message_count == 0:
        return None

    return SessionMetadata(
        session_id=session_path.stem,
        path=session_path,
        first_timestamp=first_timestamp,
        last_timestamp=last_timestamp,
        cwd=cwd,
        message_count=message_count,
        title=title,
    )


def _parse_session_full(session_id: str, session_path: Path) -> ParsedSession | None:
    """Parse a session file fully, including message conversion.

    This is the CPU-intensive operation that benefits from parallelization
    on free-threaded Python. It combines:
    - File I/O (reading JSONL)
    - JSON parsing
    - Pydantic validation
    - Message conversion

    Args:
        session_id: Session identifier
        session_path: Path to the JSONL session file

    Returns:
        ParsedSession with all data, or None if session is empty
    """
    entries = read_session(session_path)
    if not entries:
        return None

    tool_mapping = _build_tool_id_mapping(entries)
    messages: list[ChatMessage[str]] = []
    first_timestamp: datetime | None = None
    total_tokens = 0

    for entry in entries:
        msg = entry_to_chat_message(entry, session_id, tool_mapping)
        if msg is None:
            continue
        messages.append(msg)
        if first_timestamp is None and msg.timestamp:
            first_timestamp = msg.timestamp
        if msg.cost_info:
            total_tokens += msg.cost_info.token_usage.total_tokens

    if not messages:
        return None

    return ParsedSession(
        session_id=session_id,
        path=session_path,
        entries=entries,
        tool_mapping=tool_mapping,
        messages=messages,
        first_timestamp=first_timestamp,
        total_tokens=total_tokens,
    )


class ClaudeStorageProvider(StorageProvider):
    """Storage provider that reads/writes Claude Code's native format.

    Claude stores conversations as JSONL files in:
    - ~/.claude/projects/{path-encoded-project-name}/{session-id}.jsonl

    Each line is a JSON object representing a message in the conversation.

    ## Fields NOT currently used from Claude format:
    - `isSidechain`: Whether message is on a side branch
    - `userType`: Type of user ("external", etc.)
    - `cwd`: Working directory at time of message
    - `gitBranch`: Git branch at time of message
    - `version`: Claude CLI version
    - `requestId`: API request ID
    - `agentId`: Agent identifier for subagents
    - `toolUseResult`: Detailed tool result content (we extract text only)
    - `parentUuid`: Parent message for threading (we use flat history)

    ## Additional Claude data not handled:
    - `~/.claude/todos/`: Todo lists per session
    - `~/.claude/plans/`: Markdown plan files
    - `~/.claude/skills/`: Custom skills
    - `~/.claude/history.jsonl`: Command/prompt history
    """

    can_load_history = True

    def __init__(self, config: ClaudeStorageConfig | None = None) -> None:
        """Initialize Claude storage provider.

        Args:
            config: Configuration for the provider
        """
        config = config or ClaudeStorageConfig()
        super().__init__(config)
        self.base_path = Path(config.path).expanduser()
        self.projects_path = self.base_path / "projects"
        self._ensure_dirs()

    def _ensure_dirs(self) -> None:
        """Ensure required directories exist."""
        self.projects_path.mkdir(parents=True, exist_ok=True)

    def _get_project_dir(self, project_path: str) -> Path:
        """Get the directory for a project's conversations."""
        return self.projects_path / encode_project_path(project_path)

    def _find_session_path(self, session_id: str) -> Path | None:
        """Find the file path for a session by ID.

        Searches across all project directories.

        Args:
            session_id: Session ID (file stem)

        Returns:
            Path to the session file, or None if not found
        """
        for project_dir in self.projects_path.iterdir():
            if project_dir.is_dir():
                candidate = project_dir / f"{session_id}.jsonl"
                if candidate.exists():
                    return candidate
        return None

    def _list_sessions(self, project_path: str | None = None) -> list[tuple[str, Path]]:
        """List all sessions, optionally filtered by project.

        Returns:
            List of (session_id, file_path) tuples
        """
        if project_path and (project_dir := self._get_project_dir(project_path)).exists():
            return [(f.stem, f) for f in project_dir.glob("*.jsonl")]
        sessions = []
        for project_dir in self.projects_path.iterdir():
            if project_dir.is_dir():
                for f in project_dir.glob("*.jsonl"):
                    session_id = f.stem
                    sessions.append((session_id, f))
        return sessions

    def list_session_metadata(self, project_path: str | None = None) -> list[SessionMetadata]:
        """List all sessions with lightweight metadata (no full message parsing).

        This is much faster than get_sessions() as it only reads timestamps
        and message counts without validating all message content.

        Args:
            project_path: Optional project directory to filter by

        Returns:
            List of SessionMetadata objects with timestamps and counts
        """
        # return [
        #     m
        #     for _, session_path in self._list_sessions(project_path=project_path)
        #     if (m := _read_session_metadata(session_path))
        # ]
        result: list[SessionMetadata] = []
        for _, session_path in self._list_sessions(project_path=project_path):
            if metadata := _read_session_metadata(session_path):
                result.append(metadata)
        return result

    async def list_session_ids(
        self,
        pool_id: str | None = None,
        agent_name: str | None = None,
    ) -> list[str]:
        """List session IDs from Claude storage."""
        return [sid for sid, _ in self._list_sessions()]

    async def load_session(self, session_id: str) -> SessionData | None:
        """Load session data from Claude storage metadata."""
        from agentpool.sessions.models import SessionData

        session_path = self._find_session_path(session_id)
        if not session_path:
            return None
        meta = _read_session_metadata(session_path)
        if not meta:
            return None
        now = get_now()
        created_at = (
            parse_iso_timestamp(meta.first_timestamp, fallback=now) if meta.first_timestamp else now
        )
        last_active = (
            parse_iso_timestamp(meta.last_timestamp, fallback=created_at)
            if meta.last_timestamp
            else created_at
        )
        return SessionData(
            session_id=meta.session_id,
            agent_name="claude_code",
            cwd=meta.cwd or "",
            created_at=created_at,
            last_active=last_active,
            metadata={"title": meta.title, "message_count": meta.message_count}
            if meta.title
            else {"message_count": meta.message_count},
        )

    async def filter_messages(self, query: SessionQuery) -> list[ChatMessage[str]]:
        """Filter messages based on query.

        Uses parallel parsing on free-threaded Python when querying
        across multiple sessions.
        """
        # Narrow session list when a specific session is requested
        if query.name:
            path = self._find_session_path(query.name)
            sessions = [(query.name, path)] if path else []
        else:
            sessions = self._list_sessions()

        if not sessions:
            return []

        # Parse sessions in parallel (CPU-intensive)
        parsed_sessions = parallel_map(
            lambda item: _parse_session_full(item[0], item[1]),
            sessions,
        )

        # Filter messages (fast, sequential)
        messages: list[ChatMessage[str]] = []
        cutoff = query.get_time_cutoff()
        until_dt = datetime.fromisoformat(query.until) if query.until else None

        for parsed in parsed_sessions:
            if parsed is None:
                continue

            for msg in parsed.messages:
                # Apply filters
                if query.agents and msg.name not in query.agents:
                    continue

                if query.since and cutoff and msg.timestamp and msg.timestamp < cutoff:
                    continue

                if until_dt and msg.timestamp and msg.timestamp > until_dt:
                    continue

                if query.contains and query.contains not in msg.content:
                    continue

                if query.roles and msg.role not in query.roles:
                    continue

                messages.append(msg)

                if query.limit and len(messages) >= query.limit:
                    return messages

        return messages

    async def log_message(self, *, message: ChatMessage[Any]) -> None:
        """Log a message to Claude format.

        Note: session_id should be in format "project_path:session_id"
        or just "session_id" (will use default project).
        """
        raw_session_id = message.session_id or ""
        # Parse session_id to extract project path
        if ":" in raw_session_id:
            project_path, session_id = raw_session_id.split(":", 1)
        else:
            project_path = "/tmp"
            session_id = raw_session_id

        # Convert to entry and write
        entry = chat_message_to_entry(message, session_id=session_id, cwd=project_path)
        session_path = self._get_project_dir(project_path) / f"{session_id}.jsonl"
        write_entry(session_path, entry)

    async def log_session(
        self,
        *,
        session_id: str,
        node_name: str,
        start_time: datetime | None = None,
        model: str | None = None,
        agent_type: str | None = None,
    ) -> None:
        """Log a conversation start.

        In Claude format, conversations are implicit (created when first message is written).
        This is a no-op but could be extended to create an initial entry.
        """

    async def get_sessions(self, filters: QueryFilters) -> list[ConversationData]:
        """Get filtered conversations with their messages.

        Uses parallel parsing on free-threaded Python for better performance
        when processing many session files.
        """
        from agentpool_storage.models import ConversationData

        # Collect session list (lightweight - no parsing)
        sessions = self._list_sessions(project_path=filters.cwd)
        if not sessions:
            return []

        # Parse sessions in parallel (CPU-intensive Pydantic validation)
        # parallel_map automatically falls back to sequential on GIL-enabled Python
        parsed_sessions = parallel_map(
            lambda item: _parse_session_full(item[0], item[1]),
            sessions,
        )

        # Filter and build results (fast, sequential)
        result: list[ConversationData] = []
        for parsed in parsed_sessions:
            if parsed is None:
                continue

            # Apply filters
            if filters.agent_name and not any(
                m.name == filters.agent_name for m in parsed.messages
            ):
                continue
            if filters.since and parsed.first_timestamp and parsed.first_timestamp < filters.since:
                continue
            if filters.query and not any(filters.query in m.content for m in parsed.messages):
                continue

            token_usage_data: TokenUsage | None = (
                TokenUsage(total=parsed.total_tokens, prompt=0, completion=0)
                if parsed.total_tokens
                else None
            )
            conv_data = ConversationData(
                id=parsed.session_id,
                agent=parsed.messages[0].name or "claude",
                title=extract_title(parsed.path),
                start_time=(parsed.first_timestamp or get_now()).isoformat(),
                messages=parsed.messages,
                token_usage=token_usage_data,
            )

            result.append(conv_data)
            if filters.limit and len(result) >= filters.limit:
                break

        return result

    async def get_session_stats(self, filters: StatsFilters) -> dict[str, dict[str, Any]]:
        """Get conversation statistics."""
        stats: dict[str, dict[str, Any]] = defaultdict(
            lambda: {"total_tokens": 0, "messages": 0, "models": set()}
        )
        for _session_id, session_path in self._list_sessions():
            with session_path.open("r", encoding="utf-8") as f:
                for line in f:
                    stripped = line.strip()
                    if not stripped:
                        continue
                    try:
                        data = anyenv.load_json(stripped, return_type=dict)
                    except anyenv.JsonLoadError:
                        continue
                    if data.get("type") != "assistant":
                        continue
                    msg = data.get("message")
                    if not isinstance(msg, dict) or msg.get("type") != "message":
                        continue
                    model = normalize_model_name(msg.get("model", "unknown"))
                    usage = msg.get("usage", {})
                    input_tokens = usage.get("input_tokens", 0) or 0
                    output_tokens = usage.get("output_tokens", 0) or 0
                    cache_read = usage.get("cache_read_input_tokens", 0) or 0
                    total_tokens = input_tokens + output_tokens + cache_read
                    timestamp_str = data.get("timestamp", "")
                    if not timestamp_str:
                        continue
                    timestamp = parse_iso_timestamp(timestamp_str)
                    if timestamp < filters.cutoff:
                        continue
                    # Group by specified criterion
                    match filters.group_by:
                        case "model":
                            key = model or "unknown"
                        case "hour":
                            key = timestamp.strftime("%Y-%m-%d %H:00")
                        case "day":
                            key = timestamp.strftime("%Y-%m-%d")
                        case _:
                            key = "claude"  # Default agent grouping

                    stats[key]["messages"] += 1
                    stats[key]["total_tokens"] += total_tokens
                    stats[key]["models"].add(model)

        # Convert sets to lists for JSON serialization
        for value in stats.values():
            value["models"] = list(value["models"])

        return dict(stats)

    async def reset(self, *, agent_name: str | None = None, hard: bool = False) -> tuple[int, int]:
        """Reset storage.

        Warning: This will delete Claude conversation files!
        """
        conv_count = 0
        msg_count = 0
        for _session_id, session_path in self._list_sessions():
            count = count_session_messages(session_path)
            msg_count += count
            conv_count += 1

            if hard or not agent_name:
                session_path.unlink(missing_ok=True)

        return conv_count, msg_count

    async def get_session_counts(self, *, agent_name: str | None = None) -> tuple[int, int]:
        """Get counts of conversations and messages."""
        conv_count = 0
        msg_count = 0
        for _session_id, session_path in self._list_sessions():
            if count := count_session_messages(session_path):
                conv_count += 1
                msg_count += count
        return conv_count, msg_count

    async def get_session_messages(
        self,
        session_id: str,
        *,
        include_ancestors: bool = False,
    ) -> list[ChatMessage[str]]:
        """Get all messages for a session.

        Args:
            session_id: Session ID (conversation ID in Claude format)
            include_ancestors: If True, traverse parent_uuid chain to include
                messages from ancestor conversations

        Returns:
            List of messages ordered by timestamp
        """
        # Find the session file
        session_path = next((p for sid, p in self._list_sessions() if sid == session_id), None)
        if not session_path:
            return []

        # Read entries and convert to messages
        entries = read_session(session_path)
        tool_mapping = _build_tool_id_mapping(entries)
        messages = [
            m for entry in entries if (m := entry_to_chat_message(entry, session_id, tool_mapping))
        ]
        # Sort by timestamp
        now = get_now()
        messages.sort(key=lambda m: m.timestamp or now)
        if not include_ancestors or not messages:
            return messages
        # Get ancestor chain if first message has parent_id
        first_msg = messages[0]
        if first_msg.parent_id:
            ancestors = await self.get_message_ancestry(first_msg.parent_id, session_id=session_id)
            return ancestors + messages

        return messages

    async def get_message(
        self,
        message_id: str,
        *,
        session_id: str | None = None,
    ) -> ChatMessage[str] | None:
        """Get a single message by ID.

        Args:
            message_id: UUID of the message
            session_id: Optional session ID hint for faster lookup

        Returns:
            The message if found, None otherwise
        """
        # If session_id is provided, search only that session
        sessions = (
            [(session_id, p)]
            if session_id and (p := self._find_session_path(session_id))
            else self._list_sessions()
        )
        for sid, session_path in sessions:
            entries = read_session(session_path)
            tool_mapping = _build_tool_id_mapping(entries)
            for entry in entries:
                if isinstance(entry, ClaudeEntry) and entry.uuid == message_id:
                    return entry_to_chat_message(entry, sid, tool_mapping)

        return None

    async def get_message_ancestry(
        self,
        message_id: str,
        *,
        session_id: str | None = None,
    ) -> list[ChatMessage[str]]:
        """Get the ancestry chain of a message.

        Traverses parent_uuid chain to build full history.
        When session_id is provided, loads the session once and traverses in-memory.

        Args:
            message_id: UUID of the message
            session_id: Optional session ID hint for faster lookup

        Returns:
            List of messages from oldest ancestor to the specified message
        """
        # Fast path: if we know the session, load once and traverse in-memory
        if session_id and (session_path := self._find_session_path(session_id)):
            entries = read_session(session_path)
            tool_mapping = _build_tool_id_mapping(entries)
            # Build UUID -> entry index for O(1) lookups
            entry_by_uuid = {
                entry.uuid: entry
                for entry in entries
                if isinstance(entry, ClaudeUserEntry | ClaudeAssistantEntry)
            }
            ancestors: list[ChatMessage[str]] = []
            current_id: str | None = message_id
            while current_id and (found := entry_by_uuid.get(current_id)):
                if msg := entry_to_chat_message(found, session_id, tool_mapping):
                    ancestors.append(msg)
                current_id = found.parent_uuid
            ancestors.reverse()
            return ancestors

        # Slow path: search all sessions
        ancestors = []
        current_id = message_id
        while current_id:
            msg = await self.get_message(current_id)
            if not msg:
                break
            ancestors.append(msg)
            current_id = msg.parent_id
        ancestors.reverse()
        return ancestors

    async def fork_conversation(
        self,
        *,
        source_session_id: str,
        new_session_id: str,
        fork_from_message_id: str | None = None,
        new_agent_name: str | None = None,
    ) -> str | None:
        """Fork a conversation at a specific point.

        Creates a new session file. The fork point message_id is returned
        so callers can set it as parent_uuid for new messages.

        Args:
            source_session_id: Source session ID
            new_session_id: New session ID
            fork_from_message_id: UUID to fork from. If None, forks from last message
            new_agent_name: Not used in Claude format (no agent metadata in sessions)

        Returns:
            The UUID of the fork point message
        """
        # Find source session
        sessions = self._list_sessions()
        source_path = next((spath for sid, spath in sessions if sid == source_session_id), None)
        if not source_path:
            raise ValueError(f"Source conversation not found: {source_session_id}")
        # Read source entries
        entries = read_session(source_path)
        # Find fork point
        fork_point_id: str | None = None
        if fork_from_message_id:
            # Verify message exists
            found = False
            for entry in entries:
                if isinstance(entry, ClaudeEntry) and entry.uuid == fork_from_message_id:
                    found = True
                    fork_point_id = fork_from_message_id
                    break
            if not found:
                err = f"Message {fork_from_message_id} not found in conversation"
                raise ValueError(err)
        # Find last message
        elif msg_entries := [e for e in entries if isinstance(e, ClaudeEntry)]:
            fork_point_id = msg_entries[-1].uuid

        # Create new session file (empty for now - will be populated when messages added)
        # Determine project from source path structure
        project_name = source_path.parent.name
        new_path = self.projects_path / project_name / f"{new_session_id}.jsonl"
        new_path.parent.mkdir(parents=True, exist_ok=True)
        new_path.touch()
        return fork_point_id


if __name__ == "__main__":
    provider = ClaudeStorageProvider()
    print(provider._list_sessions())



"""ACP to native event converters.

This module provides conversion from ACP session updates to native agentpool
streaming events, enabling ACPAgent to yield the same event types as native agents.

This is the reverse of the conversion done in acp_server/session.py handle_event().

Also provides ACPMessageAccumulator for converting ACP notification streams
into ChatMessage objects (the reverse of replaying).
"""

from __future__ import annotations

import base64
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Literal, assert_never, overload
from uuid import uuid4

from pydantic_ai import (
    AudioUrl,
    BinaryContent,
    BinaryImage,
    CachePoint,
    DocumentUrl,
    ImageUrl,
    ModelRequest,
    ModelResponse,
    TextPart,
    ThinkingPart,
    ToolCallPart,
    ToolReturnPart,
    UserPromptPart,
    VideoUrl,
)

from acp.schema import (
    AgentMessageChunk,
    AgentPlanUpdate,
    AgentThoughtChunk,
    AudioContentBlock,
    BlobResourceContents,
    ContentToolCallContent,
    EmbeddedResourceContentBlock,
    FileEditToolCallContent,
    ImageContentBlock,
    ResourceContentBlock,
    SessionConfigSelectOption,
    TerminalToolCallContent,
    TextContentBlock,
    ToolCallProgress,
    ToolCallStart,
    UserMessageChunk,
)
from agentpool.agents.events import (
    DiffContentItem,
    LocationContentItem,
    PartDeltaEvent,
    PlanUpdateEvent,
    TerminalContentItem,
    ToolCallProgressEvent,
    ToolCallStartEvent,
)


if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

    from pydantic_ai import FinishReason, ModelMessage, ModelResponsePart, UserContent

    from acp.schema import (
        ContentBlock,
        HttpMcpServer,
        McpServer,
        SessionConfigOption,
        SessionModelState,
        SessionModeState,
        SessionUpdate,
        SseMcpServer,
        StdioMcpServer,
        StopReason,
        ToolCallContent,
        ToolCallLocation,
    )
    from agentpool.agents.events import RichAgentStreamEvent, ToolCallContentItem
    from agentpool.agents.modes import ModeCategory, ModeInfo
    from agentpool.messaging.messages import ChatMessage
    from agentpool_config.mcp_server import (
        MCPServerConfig,
        SSEMCPServerConfig,
        StdioMCPServerConfig,
        StreamableHTTPMCPServerConfig,
    )

STOP_REASON_MAP: dict[StopReason, FinishReason] = {
    "end_turn": "stop",
    "max_tokens": "length",
    "max_turn_requests": "length",
    "refusal": "content_filter",
    "cancelled": "error",
}


def get_modes(
    config_options: list[SessionConfigOption],
    available_modes: SessionModeState | None,
    available_models: SessionModelState | None,
) -> list[ModeCategory]:
    from agentpool.agents.modes import ModeCategory, ModeInfo

    categories: list[ModeCategory] = []

    if config_options:
        for config_opt in config_options:
            # Extract options from the config (ungrouped or grouped)
            mode_infos: list[ModeInfo] = []
            for i in config_opt.options:
                opts = [i] if isinstance(i, SessionConfigSelectOption) else i.options
                mode_infos.extend(
                    ModeInfo(
                        id=sub_opt.value,
                        name=sub_opt.name,
                        description=sub_opt.description or "",
                        category_id=config_opt.id,
                    )
                    for sub_opt in opts
                )

            categories.append(
                ModeCategory(
                    id=config_opt.id,
                    name=config_opt.name,
                    available_modes=mode_infos,
                    current_mode_id=config_opt.current_value,
                    category=config_opt.category or "other",
                )
            )
        return categories

    # Legacy: Convert ACP SessionModeState to ModeCategory
    if available_modes:
        modes = [
            ModeInfo(
                id=m.id,
                name=m.name,
                description=m.description or "",
                category_id="mode",
            )
            for m in available_modes.available_modes
        ]
        categories.append(
            ModeCategory(
                id="mode",
                name="Mode",
                available_modes=modes,
                current_mode_id=available_modes.current_mode_id,
                category="mode",
            )
        )

    # Legacy: Convert ACP SessionModelState to ModeCategory
    if available_models:
        models = [
            ModeInfo(
                id=m.model_id,
                name=m.name,
                description=m.description or "",
                category_id="model",
            )
            for m in available_models.available_models
        ]
        categories.append(
            ModeCategory(
                id="model",
                name="Model",
                available_modes=models,
                current_mode_id=available_models.current_model_id,
                category="model",
            )
        )

    return categories


def to_finish_reason(stop_reason: StopReason) -> FinishReason:
    return STOP_REASON_MAP.get(stop_reason, "stop")


def convert_acp_locations(
    locations: Sequence[ToolCallLocation] | None,
) -> list[LocationContentItem]:
    """Convert ACP ToolCallLocation list to native LocationContentItem list."""
    return [LocationContentItem(path=loc.path, line=loc.line) for loc in locations or []]


def convert_acp_content(content: Sequence[ToolCallContent] | None) -> list[ToolCallContentItem]:
    """Convert ACP ToolCallContent list to native ToolCallContentItem list."""
    if not content:
        return []

    result: list[ToolCallContentItem] = []
    for item in content:
        match item:
            case TerminalToolCallContent(terminal_id=terminal_id):
                result.append(TerminalContentItem(terminal_id=terminal_id))
            case FileEditToolCallContent(path=path, old_text=old_text, new_text=new_text):
                result.append(DiffContentItem(path=path, old_text=old_text, new_text=new_text))
            case ContentToolCallContent(content=TextContentBlock(text=text)):
                from agentpool.agents.events import TextContentItem

                result.append(TextContentItem(text=text))
    return result


def convert_to_acp_content(prompts: Sequence[UserContent]) -> list[ContentBlock]:
    """Convert pydantic-ai UserContent to ACP ContentBlock format.

    Handles text, images, audio, video, and document content types.

    Args:
        prompts: pydantic-ai UserContent items

    Returns:
        List of ACP ContentBlock items
    """
    content_blocks: list[ContentBlock] = []

    for item in prompts:
        match item:
            case str(text):
                content_blocks.append(TextContentBlock(text=text))

            case BinaryImage(data=data, media_type=media_type):
                encoded = base64.b64encode(data).decode("utf-8")
                content_blocks.append(ImageContentBlock(data=encoded, mime_type=media_type))

            case BinaryContent(data=data, media_type=typ):
                encoded = base64.b64encode(data).decode()
                # Handle different media types
                if item.is_image:
                    content_blocks.append(ImageContentBlock(data=encoded, mime_type=typ))
                elif item.is_audio:
                    content_blocks.append(AudioContentBlock(data=encoded, mime_type=typ))
                elif item.is_document:
                    uri = f"data:application/pdf;base64,{encoded[:50]}..."
                    blob_resource = BlobResourceContents(blob=encoded, mime_type=typ, uri=uri)
                    content_blocks.append(EmbeddedResourceContentBlock(resource=blob_resource))
                else:
                    # Generic binary as embedded resource
                    uri = f"data:{typ or 'application/octet-stream'};base64,..."
                    blob_resource = BlobResourceContents(blob=encoded, mime_type=typ, uri=uri)
                    content_blocks.append(EmbeddedResourceContentBlock(resource=blob_resource))

            case (
                AudioUrl(url=url, media_type=typ)
                | DocumentUrl(url=url, media_type=typ)
                | VideoUrl(url=url, media_type=typ)
                | ImageUrl(url=url, media_type=typ)
                # FileUrl(url=url, media_type=typ)
            ):
                name = type(item).__name__.removesuffix("Url")
                content_blocks.append(ResourceContentBlock(uri=url, name=name, mime_type=typ))

            case CachePoint():
                pass

            case _ as unreachable:
                assert_never(unreachable)

    return content_blocks


def acp_to_native_event(update: SessionUpdate) -> RichAgentStreamEvent[Any] | None:  # noqa: PLR0911
    """Convert ACP session update to native streaming event.

    Args:
        update: ACP SessionUpdate from session/update notification

    Returns:
        Corresponding native event, or None if no mapping exists
    """
    match update:
        # Text message chunks -> PartDeltaEvent with TextPartDelta
        case AgentMessageChunk(content=TextContentBlock(text=text)):
            return PartDeltaEvent.text(index=0, content=text)

        # Thought chunks -> PartDeltaEvent with ThinkingPartDelta
        case AgentThoughtChunk(content=TextContentBlock(text=text)):
            return PartDeltaEvent.thinking(index=0, content=text)

        # User message echo - usually ignored
        case UserMessageChunk():
            return None

        # Tool call start -> ToolCallStartEvent
        case ToolCallStart(
            tool_call_id=tool_call_id,
            title=title,
            kind=kind,
            content=content,
            locations=locations,
            raw_input=raw_input,
        ):
            return ToolCallStartEvent(
                tool_call_id=tool_call_id,
                tool_name=title,  # ACP uses title, not separate tool_name
                title=title,
                kind=kind or "other",
                content=convert_acp_content(list(content) if content else None),
                locations=convert_acp_locations(list(locations) if locations else None),
                raw_input=raw_input or {},
            )

        # Tool call progress -> ToolCallProgressEvent or ToolCallCompleteEvent
        case ToolCallProgress(
            tool_call_id=tool_call_id,
            status=status,
            title=title,
            content=content,
            raw_output=raw_output,
        ):
            # If completed, return ToolCallCompleteEvent for metadata injection
            if status == "completed":
                from agentpool.agents.events import ToolCallCompleteEvent

                return ToolCallCompleteEvent(
                    tool_call_id=tool_call_id,
                    tool_name=title or "unknown",
                    tool_input={},  # ACP doesn't provide input in progress updates
                    tool_result=str(raw_output) if raw_output else "",
                    agent_name="",  # Will be set by agent
                    message_id="",
                    metadata=None,  # Will be injected by agent from metadata accumulator
                )
            # Otherwise return progress event
            return ToolCallProgressEvent(
                tool_call_id=tool_call_id,
                status=status or "in_progress",
                title=title,
                items=convert_acp_content(list(content) if content else None),
                message=str(raw_output) if raw_output else None,
            )

        # Plan update -> PlanUpdateEvent
        case AgentPlanUpdate(entries=entries):
            from agentpool.utils.todos import PlanEntry

            native_entries = [
                PlanEntry(content=e.content, priority=e.priority, status=e.status) for e in entries
            ]
            return PlanUpdateEvent(entries=native_entries)

        case _:
            return None


@overload
def mcp_config_to_acp(config: StdioMCPServerConfig) -> StdioMcpServer: ...


@overload
def mcp_config_to_acp(config: SSEMCPServerConfig) -> SseMcpServer: ...


@overload
def mcp_config_to_acp(config: StreamableHTTPMCPServerConfig) -> HttpMcpServer: ...


@overload
def mcp_config_to_acp(config: MCPServerConfig) -> McpServer: ...


def mcp_config_to_acp(config: MCPServerConfig) -> McpServer:
    """Convert native MCPServerConfig to ACP McpServer format.

    If the config has tool filtering (enabled_tools or disabled_tools),
    the server is wrapped with mcp-filter proxy to apply the filtering.

    Args:
        config: agentpool MCP server configuration

    Returns:
        ACP-compatible McpServer instance, or None if conversion not possible
    """
    from acp.schema.common import EnvVariable
    from acp.schema.mcp import HttpMcpServer, SseMcpServer, StdioMcpServer
    from agentpool_config.mcp_server import (
        SSEMCPServerConfig,
        StdioMCPServerConfig,
        StreamableHTTPMCPServerConfig,
    )

    # If filtering is configured, wrap with mcp-filter first
    if config.needs_tool_filtering():
        config = config.wrap_with_mcp_filter()

    match config:
        case StdioMCPServerConfig(command=command, args=args):
            envs = [EnvVariable(name=k, value=v) for k, v in config.get_env_vars().items()]
            args = list(args) if args else []
            return StdioMcpServer(name=config.name or command, command=command, args=args, env=envs)

        case SSEMCPServerConfig(url=url):
            return SseMcpServer(name=config.name or str(url), url=url, headers=[])

        case StreamableHTTPMCPServerConfig(url=url):
            return HttpMcpServer(name=config.name or str(url), url=url)

        case _ as unreachable:
            assert_never(unreachable)


# ============================================================================
# ACP Notifications → ChatMessages Converter
# ============================================================================


@dataclass
class _PendingToolCall:
    """Tracks a tool call from start to completion."""

    tool_call_id: str
    tool_name: str
    args: dict[str, Any]
    result: Any = None
    completed: bool = False


@dataclass
class ACPMessageAccumulator:
    """Accumulates ACP session notifications into ChatMessage objects.

    This is the reverse of ACPEventConverter - instead of converting native events
    to ACP notifications for "replaying", this converts ACP notifications back
    into ChatMessage objects.

    The accumulator processes a stream of notifications and groups them by role:
    - UserMessageChunk → accumulates into ChatMessage with role="user"
    - Everything else → accumulates into ChatMessage with role="assistant"

    Message boundaries occur when the role switches.

    Example:
        ```python
        accumulator = ACPMessageAccumulator()
        for notification in acp_notifications:
            accumulator.process(notification.update)
        messages = accumulator.finalize()
        ```
    """

    session_id: str | None = None
    """Optional session ID for generated messages."""

    agent_name: str | None = None
    """Optional agent name for generated messages."""

    model_name: str | None = None
    """Optional model name for generated messages."""

    # Internal state
    _current_role: Literal["user", "assistant"] | None = field(default=None, repr=False)
    """Current role being accumulated."""

    _text_buffer: list[str] = field(default_factory=list, repr=False)
    """Accumulated text content."""

    _thinking_buffer: list[str] = field(default_factory=list, repr=False)
    """Accumulated thinking content."""

    _pending_tool_calls: dict[str, _PendingToolCall] = field(default_factory=dict, repr=False)
    """Tool calls in progress, keyed by tool_call_id."""

    _completed_tool_calls: list[_PendingToolCall] = field(default_factory=list, repr=False)
    """Tool calls that have completed."""

    _messages: list[ChatMessage[str]] = field(default_factory=list, repr=False)
    """Accumulated messages."""

    _last_parent_id: str | None = field(default=None, repr=False)
    """Parent ID for linking messages."""

    def reset(self) -> None:
        """Reset accumulator state for a new conversation."""
        self._current_role = None
        self._text_buffer.clear()
        self._thinking_buffer.clear()
        self._pending_tool_calls.clear()
        self._completed_tool_calls.clear()
        self._messages.clear()
        self._last_parent_id = None

    def _finalize_current_message(self) -> None:
        """Finalize the current message and add it to the messages list."""
        from agentpool.messaging.messages import ChatMessage

        if self._current_role is None:
            return
        message_id = str(uuid4())
        if self._current_role == "user":
            # Build user message
            content = "".join(self._text_buffer)
            parts: list[UserPromptPart] = [UserPromptPart(content=content)]
            model_message = ModelRequest(parts=parts, run_id=message_id)

            msg: ChatMessage[str] = ChatMessage(
                content=content,
                role="user",
                message_id=message_id,
                session_id=self.session_id,
                parent_id=self._last_parent_id,
                messages=[model_message],
                name=None,
            )
        else:
            # Build assistant message with proper parts
            # Structure: ModelResponse(ToolCallPart) -> ModelRequest(ToolReturnPart)
            #         -> ModelResponse(TextPart)
            all_messages: list[ModelMessage] = []
            response_parts: list[ModelResponsePart] = []
            # Add thinking part if present
            if self._thinking_buffer:
                thinking_content = "".join(self._thinking_buffer)
                response_parts.append(ThinkingPart(content=thinking_content))

            # Process completed tool calls - each creates a response + request pair
            for tc in self._completed_tool_calls:
                # First, a ModelResponse with the ToolCallPart
                call_part = ToolCallPart(
                    tool_name=tc.tool_name,
                    args=tc.args,
                    tool_call_id=tc.tool_call_id,
                )
                tool_call_response = ModelResponse(parts=[call_part], model_name=self.model_name)
                all_messages.append(tool_call_response)
                # Then, a ModelRequest with the ToolReturnPart
                return_part = ToolReturnPart(
                    tool_name=tc.tool_name,
                    content=tc.result if tc.result is not None else "",
                    tool_call_id=tc.tool_call_id,
                )
                tool_return_request = ModelRequest(parts=[return_part])
                all_messages.append(tool_return_request)

            # Add pending (incomplete) tool calls to response parts
            response_parts.extend(
                ToolCallPart(
                    tool_name=tc.tool_name,
                    args=tc.args,
                    tool_call_id=tc.tool_call_id,
                )
                for tc in self._pending_tool_calls.values()
            )

            # Add text part if present
            if self._text_buffer:
                text_content = "".join(self._text_buffer)
                response_parts.append(TextPart(content=text_content))

            # Add final response with text/thinking/pending tool calls
            if response_parts:
                final_response = ModelResponse(
                    parts=response_parts,
                    model_name=self.model_name,
                    run_id=message_id,
                )
                all_messages.append(final_response)

            msg = ChatMessage(
                content="".join(self._text_buffer),
                role="assistant",
                message_id=message_id,
                session_id=self.session_id,
                parent_id=self._last_parent_id,
                messages=all_messages,
                name=self.agent_name,
                model_name=self.model_name,
            )

        self._messages.append(msg)
        self._last_parent_id = message_id
        # Clear buffers for next message
        self._text_buffer.clear()
        self._thinking_buffer.clear()
        self._pending_tool_calls.clear()
        self._completed_tool_calls.clear()
        self._current_role = None

    def _switch_role(self, new_role: Literal["user", "assistant"]) -> None:
        """Switch to a new role, finalizing the current message if needed."""
        if self._current_role is not None and self._current_role != new_role:
            self._finalize_current_message()
        self._current_role = new_role

    def process(self, update: SessionUpdate) -> None:
        """Process a single ACP session update."""
        match update:
            # User message chunks → switch to user role
            case UserMessageChunk(content=content_block):
                self._switch_role("user")
                if isinstance(content_block, TextContentBlock):
                    self._text_buffer.append(content_block.text)

            # Agent message chunks → switch to assistant role
            case AgentMessageChunk(content=content_block):
                self._switch_role("assistant")
                if isinstance(content_block, TextContentBlock):
                    self._text_buffer.append(content_block.text)

            # Agent thought chunks → assistant role thinking
            case AgentThoughtChunk(content=content_block):
                self._switch_role("assistant")
                if isinstance(content_block, TextContentBlock):
                    self._thinking_buffer.append(content_block.text)

            # Tool call start → track pending tool call
            case ToolCallStart(tool_call_id=tool_call_id, title=title, raw_input=raw_input):
                self._switch_role("assistant")
                self._pending_tool_calls[tool_call_id] = _PendingToolCall(
                    tool_call_id=tool_call_id,
                    tool_name=title,
                    args=raw_input if isinstance(raw_input, dict) else {},
                )

            # Tool call progress → update or complete tool call
            case ToolCallProgress(tool_call_id=tool_call_id, status=status, raw_output=raw_output):
                self._switch_role("assistant")
                if tool_call_id in self._pending_tool_calls:
                    tc = self._pending_tool_calls[tool_call_id]
                    if raw_output is not None:
                        tc.result = raw_output
                    if status == "completed":
                        tc.completed = True
                        self._completed_tool_calls.append(tc)
                        del self._pending_tool_calls[tool_call_id]

            # Plan updates, mode updates, etc. - just ensure assistant role
            case AgentPlanUpdate() | _:
                if self._current_role is None:
                    self._current_role = "assistant"

    def finalize(self) -> list[ChatMessage[str]]:
        """Finalize accumulation and return all messages.

        This should be called after all notifications have been processed.
        It finalizes any pending message and returns the complete list.

        Returns:
            List of ChatMessage objects accumulated from the notifications
        """
        self._finalize_current_message()
        return list(self._messages)


def acp_notifications_to_messages(
    notifications: Iterable[SessionUpdate],
    *,
    session_id: str | None = None,
    agent_name: str | None = None,
    model_name: str | None = None,
) -> list[ChatMessage[str]]:
    """Convert a sequence of ACP notifications to ChatMessage objects.

    Convenience function that creates an accumulator, processes all notifications,
    and returns the finalized messages.

    Args:
        notifications: Iterable of SessionNotification or SessionUpdate objects
        session_id: Optional session ID for generated messages
        agent_name: Optional agent name for generated messages
        model_name: Optional model name for generated messages

    Returns:
        List of ChatMessage objects

    Example:
        ```python
        # From a list of notifications
        messages = acp_notifications_to_messages(notifications)

        # With metadata
        messages = acp_notifications_to_messages(
            notifications,
            session_id="conv-123",
            agent_name="goose",
            model_name="claude-3-opus",
        )
        ```
    """
    accumulator = ACPMessageAccumulator(
        session_id=session_id,
        agent_name=agent_name,
        model_name=model_name,
    )
    for item in notifications:
        accumulator.process(item)
    return accumulator.finalize()



"""Compaction configuration for message history management."""

from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Any, Literal

from pydantic import Field
from schemez import Schema


if TYPE_CHECKING:
    from agentpool.messaging.compaction import CompactionPipeline, CompactionStep


class FilterThinkingConfig(Schema):
    """Configuration for FilterThinking step."""

    type: Literal["filter_thinking"] = "filter_thinking"

    def build(self) -> CompactionStep:
        from agentpool.messaging.compaction import FilterThinking

        return FilterThinking()


class FilterRetryPromptsConfig(Schema):
    """Configuration for FilterRetryPrompts step."""

    type: Literal["filter_retry_prompts"] = "filter_retry_prompts"

    def build(self) -> CompactionStep:
        from agentpool.messaging.compaction import FilterRetryPrompts

        return FilterRetryPrompts()


class FilterBinaryContentConfig(Schema):
    """Configuration for FilterBinaryContent step."""

    type: Literal["filter_binary"] = "filter_binary"
    keep_references: bool = False

    def build(self) -> CompactionStep:
        from agentpool.messaging.compaction import FilterBinaryContent

        return FilterBinaryContent(keep_references=self.keep_references)


class FilterToolCallsConfig(Schema):
    """Configuration for FilterToolCalls step."""

    type: Literal["filter_tools"] = "filter_tools"
    exclude_tools: list[str] = Field(default_factory=list)
    include_only: list[str] | None = None

    def build(self) -> CompactionStep:
        from agentpool.messaging.compaction import FilterToolCalls

        return FilterToolCalls(exclude_tools=self.exclude_tools, include_only=self.include_only)


class FilterEmptyMessagesConfig(Schema):
    """Configuration for FilterEmptyMessages step."""

    type: Literal["filter_empty"] = "filter_empty"

    def build(self) -> CompactionStep:
        from agentpool.messaging.compaction import FilterEmptyMessages

        return FilterEmptyMessages()


class TruncateToolOutputsConfig(Schema):
    """Configuration for TruncateToolOutputs step."""

    type: Literal["truncate_tool_outputs"] = "truncate_tool_outputs"
    max_length: int = 2000
    suffix: str = "\n... [truncated]"

    def build(self) -> CompactionStep:
        from agentpool.messaging.compaction import TruncateToolOutputs

        return TruncateToolOutputs(max_length=self.max_length, suffix=self.suffix)


class TruncateTextPartsConfig(Schema):
    """Configuration for TruncateTextParts step."""

    type: Literal["truncate_text"] = "truncate_text"
    max_length: int = 5000
    suffix: str = "\n... [truncated]"

    def build(self) -> CompactionStep:
        from agentpool.messaging.compaction import TruncateTextParts

        return TruncateTextParts(max_length=self.max_length, suffix=self.suffix)


class KeepLastMessagesConfig(Schema):
    """Configuration for KeepLastMessages step."""

    type: Literal["keep_last"] = "keep_last"
    count: int = 10
    count_pairs: bool = True

    def build(self) -> CompactionStep:
        from agentpool.messaging.compaction import KeepLastMessages

        return KeepLastMessages(count=self.count, count_pairs=self.count_pairs)


class KeepFirstMessagesConfig(Schema):
    """Configuration for KeepFirstMessages step."""

    type: Literal["keep_first"] = "keep_first"
    count: int = 2

    def build(self) -> CompactionStep:
        from agentpool.messaging.compaction import KeepFirstMessages

        return KeepFirstMessages(count=self.count)


class KeepFirstAndLastConfig(Schema):
    """Configuration for KeepFirstAndLast step."""

    type: Literal["keep_first_last"] = "keep_first_last"
    first_count: int = 2
    last_count: int = 5

    def build(self) -> CompactionStep:
        from agentpool.messaging.compaction import KeepFirstAndLast

        return KeepFirstAndLast(first_count=self.first_count, last_count=self.last_count)


class TokenBudgetConfig(Schema):
    """Configuration for TokenBudget step."""

    type: Literal["token_budget"] = "token_budget"
    max_tokens: int = 4000
    model: str = "gpt-4o"

    def build(self) -> CompactionStep:
        from agentpool.messaging.compaction import TokenBudget

        return TokenBudget(max_tokens=self.max_tokens, model=self.model)


class SummarizeConfig(Schema):
    """Configuration for Summarize step."""

    type: Literal["summarize"] = "summarize"
    model: str = "openai:gpt-4o-mini"
    threshold: int = 15
    keep_recent: int = 5
    summary_prompt: str | None = None

    def build(self) -> CompactionStep:
        from agentpool.messaging.compaction import Summarize

        kwargs: dict[str, Any] = {
            "model": self.model,
            "threshold": self.threshold,
            "keep_recent": self.keep_recent,
        }
        if self.summary_prompt:
            kwargs["summary_prompt"] = self.summary_prompt
        return Summarize(**kwargs)


class WhenMessageCountExceedsConfig(Schema):
    """Configuration for WhenMessageCountExceeds wrapper."""

    type: Literal["when_count_exceeds"] = "when_count_exceeds"
    threshold: int = 20
    step: "CompactionStepConfig"

    def build(self) -> CompactionStep:
        from agentpool.messaging.compaction import WhenMessageCountExceeds

        return WhenMessageCountExceeds(step=self.step.build(), threshold=self.threshold)


# Union of all config types with discriminator
CompactionStepConfig = Annotated[
    FilterThinkingConfig
    | FilterRetryPromptsConfig
    | FilterBinaryContentConfig
    | FilterToolCallsConfig
    | FilterEmptyMessagesConfig
    | TruncateToolOutputsConfig
    | TruncateTextPartsConfig
    | KeepLastMessagesConfig
    | KeepFirstMessagesConfig
    | KeepFirstAndLastConfig
    | TokenBudgetConfig
    | SummarizeConfig
    | WhenMessageCountExceedsConfig,
    Field(discriminator="type"),
]

# Update forward reference
WhenMessageCountExceedsConfig.model_rebuild()


class CompactionConfig(Schema):
    """Configuration for message compaction/summarization.

    Example YAML:
        ```yaml
        compaction:
          steps:
            - type: filter_thinking
            - type: truncate_tool_outputs
              max_length: 1000
            - type: keep_last
              count: 10
        ```

    Or use a preset:
        ```yaml
        compaction:
          preset: balanced
        ```
    """

    preset: Literal["minimal", "balanced", "summarizing"] | None = None
    """Use a predefined compaction pipeline.

    - minimal: Aggressively minimize context (filter thinking, truncate, keep last 10)
    - balanced: Balance context preservation with size (filter thinking, truncate,
                keep first 2 + last 8)
    - summarizing: Summarize old messages (filter thinking, summarize when > 20 messages)
    """

    steps: list[CompactionStepConfig] = Field(default_factory=list)
    """Ordered list of compaction steps to apply (ignored if preset is set)."""

    def build(self) -> CompactionPipeline:
        """Build a CompactionPipeline from this configuration."""
        from agentpool.messaging.compaction import (
            CompactionPipeline,
            balanced_context,
            minimal_context,
            summarizing_context,
        )

        if self.preset:
            match self.preset:
                case "minimal":
                    return minimal_context()
                case "balanced":
                    return balanced_context()
                case "summarizing":
                    return summarizing_context()

        return CompactionPipeline(steps=[step.build() for step in self.steps])



"""Models for agentpool standalone tool configuration."""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from exxec_config import ExecutionEnvironmentConfig  # noqa: TC002
from pydantic import ConfigDict, Field

from agentpool_config.converters import ConversionConfig  # noqa: TC001
from agentpool_config.tools import BaseToolConfig


if TYPE_CHECKING:
    from agentpool.tools.base import Tool


class BashToolConfig(BaseToolConfig):
    """Configuration for bash command execution tool.

    Example:
        ```yaml
        tools:
          - type: bash
            timeout: 30.0
            output_limit: 10000
            requires_confirmation: true
            environment:
              type: mock
              deterministic_ids: true
        ```
    """

    model_config = ConfigDict(title="Bash Tool")

    type: Literal["bash"] = Field("bash", init=False)
    """Bash command execution tool."""

    timeout: float | None = Field(
        default=None,
        examples=[30.0, 60.0, 120.0],
        title="Command timeout",
    )
    """Command timeout in seconds. None means no timeout."""

    output_limit: int | None = Field(
        default=None,
        examples=[10000, 50000, 100000],
        title="Output limit",
    )
    """Maximum bytes of output to return."""

    environment: ExecutionEnvironmentConfig | None = Field(
        default=None,
        title="Execution environment",
    )
    """Execution environment for command execution. Falls back to agent's env if not set."""

    def get_tool(self) -> Tool:
        """Convert config to BashTool instance."""
        from agentpool.tool_impls.bash import create_bash_tool

        env = self.environment.get_provider() if self.environment else None
        return create_bash_tool(
            env=env,
            timeout=self.timeout,
            output_limit=self.output_limit,
            name=self.name or "bash",
            description=self.description or "Execute a shell command and return the output.",
            requires_confirmation=self.requires_confirmation,
        )


class AgentCliToolConfig(BaseToolConfig):
    """Configuration for agent CLI tool.

    Example:
        ```yaml
        tools:
          - type: agent_cli
        ```
    """

    model_config = ConfigDict(title="Agent CLI Tool")

    type: Literal["agent_cli"] = Field("agent_cli", init=False)
    """Agent CLI tool."""

    def get_tool(self) -> Tool:
        """Convert config to AgentCliTool instance."""
        from agentpool.tool_impls.agent_cli import create_agent_cli_tool

        return create_agent_cli_tool(
            name=self.name or "run_agent_cli_command",
            description=self.description or "Execute an internal agent management command.",
            requires_confirmation=self.requires_confirmation,
        )


class QuestionToolConfig(BaseToolConfig):
    """Configuration for user interaction tool.

    Example:
        ```yaml
        tools:
          - type: question
        ```
    """

    model_config = ConfigDict(title="Ask User Tool")

    type: Literal["question"] = Field("question", init=False)
    """User interaction tool."""

    def get_tool(self) -> Tool:
        """Convert config to QuestionTool instance."""
        from agentpool.tool_impls.question import create_question_tool

        return create_question_tool(
            name=self.name or "question",
            description=self.description or "Ask the user a clarifying question.",
            requires_confirmation=self.requires_confirmation,
        )


class ExecuteCodeToolConfig(BaseToolConfig):
    """Configuration for Python code execution tool.

    Example:
        ```yaml
        tools:
          - type: execute_code
            requires_confirmation: true
            environment:
              type: mock
              deterministic_ids: true
        ```
    """

    model_config = ConfigDict(title="Execute Code Tool")

    type: Literal["execute_code"] = Field("execute_code", init=False)
    """Python code execution tool."""

    environment: ExecutionEnvironmentConfig | None = Field(
        default=None,
        title="Execution environment",
    )
    """Execution environment for code execution. Falls back to agent's env if not set."""

    def get_tool(self) -> Tool:
        """Convert config to ExecuteCodeTool instance."""
        from agentpool.tool_impls.execute_code import create_execute_code_tool

        env = self.environment.get_provider() if self.environment else None
        return create_execute_code_tool(
            env=env,
            name=self.name or "execute_code",
            description=self.description or "Execute Python code and return the result.",
            requires_confirmation=self.requires_confirmation,
        )


class ReadToolConfig(BaseToolConfig):
    """Configuration for file reading tool.

    Example:
        ```yaml
        tools:
          - type: read
            max_file_size_kb: 128
            max_image_size: 1500
            large_file_tokens: 10000
            conversion:
              default_provider: markitdown
            environment:
              type: local
        ```
    """

    model_config = ConfigDict(title="Read Tool")

    type: Literal["read"] = Field("read", init=False)
    """File reading tool."""

    environment: ExecutionEnvironmentConfig | None = Field(
        default=None,
        title="Execution environment",
    )
    """Execution environment for filesystem access. Falls back to agent's env if not set."""

    cwd: str | None = Field(
        default=None,
        title="Working directory",
    )
    """Working directory for resolving relative paths."""

    max_file_size_kb: int = Field(
        default=64,
        examples=[64, 128, 256],
        title="Max file size",
    )
    """Maximum file size in KB for read operations."""

    max_image_size: int | None = Field(
        default=2000,
        examples=[1500, 2000, 2500],
        title="Max image dimensions",
    )
    """Max width/height for images in pixels. Images are auto-resized if larger."""

    max_image_bytes: int | None = Field(
        default=None,
        title="Max image file size",
    )
    """Max file size for images in bytes. Images are compressed if larger."""

    large_file_tokens: int = Field(
        default=12_000,
        examples=[10_000, 12_000, 15_000],
        title="Large file threshold",
    )
    """Token threshold for switching to structure map for large files."""

    map_max_tokens: int = Field(
        default=2048,
        examples=[1024, 2048, 4096],
        title="Structure map max tokens",
    )
    """Maximum tokens for structure map output."""

    conversion: ConversionConfig | None = Field(
        default=None,
        title="Conversion config",
    )
    """Optional conversion config for binary files. If set, converts supported files to markdown."""

    def get_tool(self) -> Tool:
        """Convert config to ReadTool instance."""
        from agentpool.tool_impls.read import create_read_tool

        env = self.environment.get_provider() if self.environment else None

        # Create converter if conversion config is provided
        converter = None
        if self.conversion is not None:
            try:
                from agentpool.prompts.conversion_manager import ConversionManager

                converter = ConversionManager(self.conversion)
            except Exception:  # noqa: BLE001
                # ConversionManager not available, continue without it
                pass

        return create_read_tool(
            env=env,
            converter=converter,
            cwd=self.cwd,
            max_file_size_kb=self.max_file_size_kb,
            max_image_size=self.max_image_size,
            max_image_bytes=self.max_image_bytes,
            large_file_tokens=self.large_file_tokens,
            map_max_tokens=self.map_max_tokens,
            name=self.name or "read",
            description=self.description or "Read file contents with automatic format detection.",
            requires_confirmation=self.requires_confirmation,
        )


class ListDirectoryToolConfig(BaseToolConfig):
    """Configuration for directory listing tool.

    Example:
        ```yaml
        tools:
          - type: list_directory
            max_items: 1000
            environment:
              type: local
        ```
    """

    model_config = ConfigDict(title="List Directory Tool")

    type: Literal["list_directory"] = Field("list_directory", init=False)
    """Directory listing tool."""

    environment: ExecutionEnvironmentConfig | None = Field(
        default=None,
        title="Execution environment",
    )
    """Execution environment for filesystem access. Falls back to agent's env if not set."""

    cwd: str | None = Field(
        default=None,
        title="Working directory",
    )
    """Working directory for resolving relative paths."""

    max_items: int = Field(
        default=500,
        examples=[500, 1000, 2000],
        title="Max items",
    )
    """Maximum number of items to return (safety limit)."""

    def get_tool(self) -> Tool:
        """Convert config to ListDirectoryTool instance."""
        from agentpool.tool_impls.list_directory import create_list_directory_tool

        env = self.environment.get_provider() if self.environment else None
        return create_list_directory_tool(
            env=env,
            cwd=self.cwd,
            max_items=self.max_items,
            name=self.name or "list_directory",
            description=self.description or "List files in a directory with filtering support.",
            requires_confirmation=self.requires_confirmation,
        )


class GrepToolConfig(BaseToolConfig):
    """Configuration for grep search tool.

    Example:
        ```yaml
        tools:
          - type: grep
            max_output_kb: 128
            use_subprocess_grep: true
            environment:
              type: local
        ```
    """

    model_config = ConfigDict(title="Grep Tool")

    type: Literal["grep"] = Field("grep", init=False)
    """Grep search tool."""

    environment: ExecutionEnvironmentConfig | None = Field(
        default=None,
        title="Execution environment",
    )
    """Execution environment for filesystem access. Falls back to agent's env if not set."""

    cwd: str | None = Field(
        default=None,
        title="Working directory",
    )
    """Working directory for resolving relative paths."""

    max_output_kb: int = Field(
        default=64,
        examples=[64, 128, 256],
        title="Max output size",
    )
    """Maximum output size in KB."""

    use_subprocess_grep: bool = Field(
        default=True,
        title="Use subprocess grep",
    )
    """Use ripgrep/grep subprocess if available (faster for large codebases)."""

    def get_tool(self) -> Tool:
        """Convert config to GrepTool instance."""
        from agentpool.tool_impls.grep import create_grep_tool

        env = self.environment.get_provider() if self.environment else None
        return create_grep_tool(
            env=env,
            cwd=self.cwd,
            max_output_kb=self.max_output_kb,
            use_subprocess_grep=self.use_subprocess_grep,
            name=self.name or "grep",
            description=self.description or "Search file contents for patterns.",
            requires_confirmation=self.requires_confirmation,
        )


class DeletePathToolConfig(BaseToolConfig):
    """Configuration for delete path tool.

    Example:
        ```yaml
        tools:
          - type: delete_path
            requires_confirmation: true
            environment:
              type: local
        ```
    """

    model_config = ConfigDict(title="Delete Path Tool")

    type: Literal["delete_path"] = Field("delete_path", init=False)
    """Delete path tool."""

    environment: ExecutionEnvironmentConfig | None = Field(
        default=None,
        title="Execution environment",
    )
    """Execution environment for filesystem access. Falls back to agent's env if not set."""

    cwd: str | None = Field(
        default=None,
        title="Working directory",
    )
    """Working directory for resolving relative paths."""

    def get_tool(self) -> Tool:
        """Convert config to DeletePathTool instance."""
        from agentpool.tool_impls.delete_path import create_delete_path_tool

        env = self.environment.get_provider() if self.environment else None
        return create_delete_path_tool(
            env=env,
            cwd=self.cwd,
            name=self.name or "delete_path",
            description=self.description or "Delete a file or directory.",
            requires_confirmation=self.requires_confirmation,
        )


class DownloadFileToolConfig(BaseToolConfig):
    """Configuration for file download tool.

    Example:
        ```yaml
        tools:
          - type: download_file
            chunk_size: 16384
            timeout: 60.0
            environment:
              type: local
        ```
    """

    model_config = ConfigDict(title="Download File Tool")

    type: Literal["download_file"] = Field("download_file", init=False)
    """File download tool."""

    environment: ExecutionEnvironmentConfig | None = Field(
        default=None,
        title="Execution environment",
    )
    """Execution environment for filesystem access. Falls back to agent's env if not set."""

    cwd: str | None = Field(
        default=None,
        title="Working directory",
    )
    """Working directory for resolving relative paths."""

    chunk_size: int = Field(
        default=8192,
        examples=[8192, 16384, 32768],
        title="Chunk size",
    )
    """Size of chunks to download in bytes."""

    timeout: float = Field(
        default=30.0,
        examples=[30.0, 60.0, 120.0],
        title="Request timeout",
    )
    """Request timeout in seconds."""

    def get_tool(self) -> Tool:
        """Convert config to DownloadFileTool instance."""
        from agentpool.tool_impls.download_file import create_download_file_tool

        env = self.environment.get_provider() if self.environment else None
        return create_download_file_tool(
            env=env,
            cwd=self.cwd,
            chunk_size=self.chunk_size,
            timeout=self.timeout,
            name=self.name or "download_file",
            description=self.description or "Download a file from a URL.",
            requires_confirmation=self.requires_confirmation,
        )


# Union type for agentpool tool configs
AgentpoolToolConfig = (
    AgentCliToolConfig
    | QuestionToolConfig
    | BashToolConfig
    | DeletePathToolConfig
    | DownloadFileToolConfig
    | ExecuteCodeToolConfig
    | GrepToolConfig
    | ListDirectoryToolConfig
    | ReadToolConfig
)



from __future__ import annotations

from typing import Any

from codex_adapter.models.base import CodexBaseModel
from codex_adapter.models.codex_types import ModelRerouteReason  # noqa: TC001
from codex_adapter.models.misc import (  # noqa: TC001
    AppInfo,
    RateLimitSnapshot,
    TextRange,
    Thread,
    Turn,
    TurnError,
    TurnPlanStep,
)
from codex_adapter.models.thread_item import ThreadItem  # noqa: TC001
from codex_adapter.models.thread_status import ThreadStatusValue  # noqa: TC001
from codex_adapter.models.token_usage import ThreadTokenUsage  # noqa: TC001


class TurnPlanUpdatedData(CodexBaseModel):
    """Payload for turn/plan/updated notification."""

    thread_id: str
    turn_id: str
    explanation: str | None = None
    plan: list[TurnPlanStep]


# Item lifecycle notifications


class ItemStartedData(CodexBaseModel):
    """Payload for item/started notification (V2 protocol)."""

    thread_id: str
    turn_id: str
    item: ThreadItem


class ItemCompletedData(CodexBaseModel):
    """Payload for item/completed notification (V2 protocol)."""

    thread_id: str
    turn_id: str
    item: ThreadItem


class RawResponseItemCompletedData(CodexBaseModel):
    """Payload for rawResponseItem/completed notification."""

    thread_id: str
    turn_id: str
    item: ThreadItem


# Item delta notifications


class AgentMessageDeltaData(CodexBaseModel):
    """Payload for item/agentMessage/delta notification."""

    thread_id: str
    turn_id: str
    item_id: str
    delta: str


class PlanDeltaData(CodexBaseModel):
    """Payload for item/plan/delta notification."""

    thread_id: str
    turn_id: str
    item_id: str
    delta: str


class ReasoningTextDeltaData(CodexBaseModel):
    """Payload for item/reasoning/textDelta notification."""

    thread_id: str
    turn_id: str
    item_id: str
    delta: str
    content_index: int


class ReasoningSummaryTextDeltaData(CodexBaseModel):
    """Payload for item/reasoning/summaryTextDelta notification."""

    thread_id: str
    turn_id: str
    item_id: str
    delta: str
    summary_index: int


class ReasoningSummaryPartAddedData(CodexBaseModel):
    """Payload for item/reasoning/summaryPartAdded notification."""

    thread_id: str
    turn_id: str
    item_id: str
    summary_index: int


class CommandExecutionOutputDeltaData(CodexBaseModel):
    """Payload for item/commandExecution/outputDelta notification."""

    thread_id: str
    turn_id: str
    item_id: str
    delta: str


class CommandExecutionTerminalInteractionData(CodexBaseModel):
    """Payload for item/commandExecution/terminalInteraction notification."""

    thread_id: str
    turn_id: str
    item_id: str
    process_id: str
    stdin: str


class FileChangeOutputDeltaData(CodexBaseModel):
    """Payload for item/fileChange/outputDelta notification."""

    thread_id: str
    turn_id: str
    item_id: str
    delta: str


class McpToolCallProgressData(CodexBaseModel):
    """Payload for item/mcpToolCall/progress notification."""

    thread_id: str
    turn_id: str
    item_id: str
    message: str


# MCP/Account/System notifications


class McpServerOAuthLoginCompletedData(CodexBaseModel):
    """Payload for mcpServer/oauthLogin/completed notification."""

    name: str
    success: bool
    error: str | None = None


class ThreadStartedData(CodexBaseModel):
    """Payload for thread/started notification (V2 protocol)."""

    thread: Thread

    @property
    def thread_id(self) -> str:
        """Thread ID derived from the thread object."""
        return self.thread.id


class ThreadStatusChangedData(CodexBaseModel):
    """Payload for thread/status/changed notification."""

    thread_id: str
    status: ThreadStatusValue


class ThreadArchivedData(CodexBaseModel):
    """Payload for thread/archived notification."""

    thread_id: str


class ThreadUnarchivedData(CodexBaseModel):
    """Payload for thread/unarchived notification."""

    thread_id: str


class ThreadNameUpdatedData(CodexBaseModel):
    """Payload for thread/name/updated notification."""

    thread_id: str
    thread_name: str | None = None


class ThreadTokenUsageUpdatedData(CodexBaseModel):
    """Payload for thread/tokenUsage/updated notification (V2 protocol)."""

    thread_id: str
    turn_id: str
    token_usage: ThreadTokenUsage


class ThreadCompactedData(CodexBaseModel):
    """Payload for thread/compacted notification."""

    thread_id: str
    turn_id: str | None = None


# Turn lifecycle notifications


class TurnStartedData(CodexBaseModel):
    """Payload for turn/started notification (V2 protocol)."""

    thread_id: str
    turn: Turn


class TurnCompletedData(CodexBaseModel):
    """Payload for turn/completed notification (V2 protocol)."""

    thread_id: str
    turn: Turn


class TurnErrorData(CodexBaseModel):
    """Payload for turn/error notification."""

    thread_id: str
    turn_id: str
    error: str


class TurnDiffUpdatedData(CodexBaseModel):
    """Payload for turn/diff/updated notification."""

    thread_id: str
    turn_id: str
    diff: str


class AccountRateLimitsUpdatedData(CodexBaseModel):
    """Payload for account/rateLimits/updated notification."""

    rate_limits: RateLimitSnapshot


class AccountLoginCompletedData(CodexBaseModel):
    """Payload for account/login/completed notification."""

    login_id: str | None = None
    success: bool
    error: str | None = None


class AuthStatusChangeData(CodexBaseModel):
    """Payload for authStatusChange notification (legacy v1)."""

    status: str


class LoginChatGptCompleteData(CodexBaseModel):
    """Payload for loginChatGptComplete notification (legacy v1)."""

    success: bool


class SessionConfiguredData(CodexBaseModel):
    """Payload for sessionConfigured notification."""

    config: dict[str, Any]  # Session config - flexible structure


class DeprecationNoticeData(CodexBaseModel):
    """Payload for deprecationNotice notification."""

    summary: str
    details: str | None = None


class WindowsWorldWritableWarningData(CodexBaseModel):
    """Payload for windows/worldWritableWarning notification."""

    sample_paths: list[str]
    extra_count: int
    failed_scan: bool


class ErrorEventData(CodexBaseModel):
    """Payload for error event."""

    error: TurnError
    will_retry: bool
    thread_id: str
    turn_id: str


class ModelReroutedData(CodexBaseModel):
    """Payload for model/rerouted notification."""

    thread_id: str
    turn_id: str
    from_model: str
    to_model: str
    reason: ModelRerouteReason


class ConfigWarningData(CodexBaseModel):
    """Payload for configWarning notification."""

    summary: str
    details: str | None = None
    path: str | None = None
    range: TextRange | None = None


class AppListUpdatedData(CodexBaseModel):
    """Payload for app/list/updated notification."""

    data: list[AppInfo]


class ContextCompactedData(CodexBaseModel):
    """Payload for thread/compacted/v2 notification."""

    thread_id: str
    turn_id: str | None = None


class ServerRequestResolvedData(CodexBaseModel):
    """Payload for serverRequest/resolved notification."""

    thread_id: str
    request_id: int | str


class AccountUpdatedData(CodexBaseModel):
    """Payload for account/updated notification."""

    auth_mode: str | None = None


# Union type of all event data
EventData = (
    # Thread lifecycle
    ThreadStartedData
    | ThreadStatusChangedData
    | ThreadArchivedData
    | ThreadUnarchivedData
    | ThreadNameUpdatedData
    | ThreadTokenUsageUpdatedData
    | ThreadCompactedData
    # Turn lifecycle
    | TurnStartedData
    | TurnCompletedData
    | TurnErrorData
    | TurnDiffUpdatedData
    | TurnPlanUpdatedData
    # Item lifecycle
    | ItemStartedData
    | ItemCompletedData
    | RawResponseItemCompletedData
    # Item deltas - agent messages
    | AgentMessageDeltaData
    # Item deltas - plan
    | PlanDeltaData
    # Item deltas - reasoning
    | ReasoningTextDeltaData
    | ReasoningSummaryTextDeltaData
    | ReasoningSummaryPartAddedData
    # Item deltas - command execution
    | CommandExecutionOutputDeltaData
    | CommandExecutionTerminalInteractionData
    # Item deltas - file changes
    | FileChangeOutputDeltaData
    # Item deltas - MCP tool calls
    | McpToolCallProgressData
    # MCP OAuth
    | McpServerOAuthLoginCompletedData
    # Account/Auth events
    | AccountUpdatedData
    | AccountRateLimitsUpdatedData
    | AccountLoginCompletedData
    | AuthStatusChangeData
    | LoginChatGptCompleteData
    # System events
    | SessionConfiguredData
    | DeprecationNoticeData
    | WindowsWorldWritableWarningData
    # Error events
    | ErrorEventData
    # New events
    | ModelReroutedData
    | ConfigWarningData
    | AppListUpdatedData
    | ContextCompactedData
    | ServerRequestResolvedData
)



"""Zed IDE storage format models."""

from __future__ import annotations

import io
from typing import Any, Literal

import anyenv
from pydantic import AliasChoices, BaseModel, ConfigDict, Field


class ZedBaseModel(BaseModel):
    """Base model with Zed storage."""

    model_config = ConfigDict(use_attribute_docstrings=True, extra="forbid")


class ZedMentionUri(ZedBaseModel):
    """Mention URI - can be File, Directory, Symbol, etc."""

    File: dict[str, Any] | None = None
    Directory: dict[str, Any] | None = None
    Symbol: dict[str, Any] | None = None
    Selection: dict[str, Any] | None = None
    Thread: dict[str, Any] | None = None
    TextThread: dict[str, Any] | None = None
    Rule: dict[str, Any] | None = None
    Fetch: dict[str, Any] | None = None
    PastedImage: bool | None = None


class ZedMention(ZedBaseModel):
    """A file/symbol mention in Zed."""

    uri: ZedMentionUri
    content: str


class ZedImage(ZedBaseModel):
    """An image in Zed (base64 encoded)."""

    source: str  # base64 encoded


class ZedThinking(ZedBaseModel):
    """Thinking block from model."""

    text: str
    signature: str | None = None


class ZedToolUse(ZedBaseModel):
    """Tool use block."""

    id: str
    name: str
    raw_input: str
    input: dict[str, Any]
    is_input_complete: bool = True
    thought_signature: str | None = None


class ZedToolResult(ZedBaseModel):
    """Tool result."""

    tool_use_id: str
    tool_name: str
    is_error: bool = False
    content: dict[str, Any] | str | None = None
    output: dict[str, Any] | str | None = None


# v0.2.0+ nested message format


class ZedUserMessage(ZedBaseModel):
    """User message in Zed thread (v0.2.0+ format)."""

    id: str
    content: list[dict[str, Any]]  # Can contain Text, Image, Mention


class ZedAgentMessage(ZedBaseModel):
    """Agent message in Zed thread (v0.2.0+ format)."""

    content: list[dict[str, Any]]  # Can contain Text, Thinking, ToolUse
    tool_results: dict[str, ZedToolResult] = Field(default_factory=dict)
    reasoning_details: Any | None = None


class ZedNestedMessage(ZedBaseModel):
    """A message in Zed thread v0.2.0+ - nested under User or Agent key."""

    User: ZedUserMessage | None = None
    Agent: ZedAgentMessage | None = None


# Flat message format (v0.1.0, v0.2.0)


class ZedSegment(ZedBaseModel):
    """A segment in a flat message."""

    type: Literal["text", "thinking"]
    text: str | None = None
    signature: str | None = None  # For thinking segments


class ZedFlatMessage(ZedBaseModel):
    """A message in flat format (used in v0.1.0 and v0.2.0)."""

    id: int
    role: Literal["user", "assistant"]
    segments: list[ZedSegment] = Field(default_factory=list)
    tool_uses: list[dict[str, Any]] = Field(default_factory=list)
    tool_results: list[dict[str, Any]] = Field(default_factory=list)
    context: str = ""
    creases: list[Any] = Field(default_factory=list)
    is_hidden: bool = False


# Union of all message formats - put ZedFlatMessage first since it's more specific
# (has required 'id' and 'role' fields that ZedNestedMessage doesn't have)
ZedMessage = ZedFlatMessage | ZedNestedMessage


class ZedLanguageModel(ZedBaseModel):
    """Model configuration."""

    provider: str
    model: str


class ZedTokenUsage(ZedBaseModel):
    """Token usage for a request or cumulative."""

    input_tokens: int = 0
    output_tokens: int = 0
    cache_creation_input_tokens: int = 0
    cache_read_input_tokens: int = 0


class ZedWorktreeSnapshot(ZedBaseModel):
    """Git worktree snapshot."""

    worktree_path: str
    git_state: dict[str, Any] | None = None


class ZedProjectSnapshot(ZedBaseModel):
    """Project snapshot with git state."""

    worktree_snapshots: list[ZedWorktreeSnapshot] = Field(default_factory=list)
    unsaved_buffer_paths: list[str] = Field(default_factory=list)
    timestamp: str | None = None


class ZedThread(ZedBaseModel):
    """A Zed conversation thread."""

    model_config = {"populate_by_name": True}

    # v0.3.0 uses "title", v0.2.0 uses "summary"
    title: str = Field(alias="title", validation_alias=AliasChoices("title", "summary"))
    messages: list[ZedMessage | Literal["Resume"]]  # Control messages
    updated_at: str
    version: str | None = None
    detailed_summary: str | None = None  # v0.3.0 field
    detailed_summary_state: str | dict[str, Any] | None = None  # v0.2.0 field
    initial_project_snapshot: ZedProjectSnapshot | None = None
    cumulative_token_usage: ZedTokenUsage = Field(default_factory=ZedTokenUsage)
    # v0.2.0: list of token usage per request
    # v0.3.0+: dict keyed by message ID
    request_token_usage: list[ZedTokenUsage] | dict[str, ZedTokenUsage] = Field(
        default_factory=list
    )
    model: ZedLanguageModel | None = None
    completion_mode: str | None = None
    profile: str | None = None
    exceeded_window_error: Any | None = None
    tool_use_limit_reached: bool = False
    imported: bool = False  # Whether thread was imported from another source

    @classmethod
    def from_compressed(cls, data: bytes, data_type: Literal["zstd", "plain"]) -> ZedThread:
        """Decompress and parse thread data.

        Args:
            data: Compressed thread data
            data_type: Type of compression ("zstd" or "plain")

        Returns:
            Parsed ZedThread object
        """
        import zstandard

        if data_type == "zstd":
            dctx = zstandard.ZstdDecompressor()
            reader = dctx.stream_reader(io.BytesIO(data))
            json_data = reader.read()
        else:
            json_data = data

        thread_dict = anyenv.load_json(json_data)
        return cls.model_validate(thread_dict)



"""Base class for component registries."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import MutableMapping
from typing import TYPE_CHECKING, Any, TypeVar

from psygnal.containers import EventedDict


if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Iterator, Sequence

    from psygnal.containers import DictEvents


TKey = TypeVar("TKey", str, int)


class AgentPoolError(Exception):
    """Base exception for all AgentPool errors."""


class BaseRegistry[TKey, TItem](MutableMapping[TKey, TItem], ABC):
    """Base class for registries providing item storage and change notifications.

    This registry implements a dictionary-like interface backed by an EventedDict,
    providing automatic event emission for all mutations (additions, removals,
    modifications).

    Features:
    - Dictionary-like access (registry[key] = item)
    - Event emission for all changes
    - Item validation
    - Type safety
    - Customizable error handling

    Available events (accessed via .events):
        - adding(key, value): Before an item is added
        - added(key, value): After an item is added
        - removing(key, value): Before an item is removed
        - removed(key, value): After an item is removed
        - changing(key, value): Before an item is modified
        - changed(key, value): After an item is modified

    To implement, override:
    - _validate_item: Custom validation/transformation of items
    - _error_class: Custom error type for exceptions
    """

    def __init__(self) -> None:
        """Initialize an empty registry."""
        self._items = EventedDict[TKey, TItem]()
        self._initialized = False
        self._configs: dict[TKey, Any] = {}

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._items})"

    @property
    def is_empty(self) -> bool:
        """Check if registry has any items."""
        return not bool(self._items)

    def has_item(self, key: TKey) -> bool:
        """Check if an item is registered."""
        return key in self._items

    @property
    def events(self) -> DictEvents:
        """Access to all dictionary events."""
        return self._items.events

    def register(self, key: TKey, item: TItem | Any, replace: bool = False) -> None:
        """Register an item."""
        if key in self._items and not replace:
            msg = f"Item already registered: {key}"
            raise self._error_class(msg)

        validated_item = self._validate_item(item)
        self._items[key] = validated_item

    def get(self, key: TKey) -> TItem:  # type: ignore
        """Get an item by key."""
        return self[key]

    def list_items(self) -> Sequence[TKey]:
        """List all registered item keys."""
        return list(self._items.keys())

    def reset(self) -> None:
        """Reset registry to initial state."""
        self._items.clear()
        self._configs.clear()
        self._initialized = False

    async def startup(self) -> None:
        """Initialize all registered items."""
        if self._initialized:
            return

        try:
            for item in self._items.values():
                await self._initialize_item(item)
            self._initialized = True
        except Exception as exc:
            await self.shutdown()
            msg = f"Registry startup failed: {exc}"
            raise self._error_class(msg) from exc

    async def shutdown(self) -> None:
        """Cleanup all registered items."""
        if not self._initialized:
            return

        errors: list[tuple[TKey, Exception]] = []

        for key, item in self._items.items():
            try:
                await self._cleanup_item(item)
            except Exception as exc:  # noqa: BLE001
                errors.append((key, exc))

        self._initialized = False

        if errors:
            error_msgs = [f"{key}: {exc}" for key, exc in errors]
            msg = f"Errors during shutdown: {', '.join(error_msgs)}"
            raise self._error_class(msg)

    @property
    def _error_class(self) -> type[AgentPoolError]:
        """Error class to use for this registry."""
        return AgentPoolError

    @abstractmethod
    def _validate_item(self, item: Any) -> TItem:
        """Validate and possibly transform item before registration."""

    async def _initialize_item(self, item: TItem) -> None:
        """Initialize an item during startup."""
        if hasattr(item, "startup") and callable(item.startup):  # pyright: ignore
            await item.startup()  # pyright: ignore  # ty: ignore

    async def _cleanup_item(self, item: TItem) -> None:
        """Clean up an item during shutdown."""
        if hasattr(item, "shutdown") and callable(item.shutdown):  # pyright: ignore
            await item.shutdown()  # pyright: ignore  # ty: ignore

    # Implementing MutableMapping methods
    def __getitem__(self, key: TKey) -> TItem:
        try:
            return self._items[key]
        except KeyError as exc:
            raise self._error_class(f"Item not found: {key}") from exc

    def __setitem__(self, key: TKey, value: Any) -> None:
        """Support dict-style assignment."""
        self.register(key, value)

    def __contains__(self, key: object) -> bool:
        """Support 'in' operator without raising exceptions."""
        return key in self._items

    def __delitem__(self, key: TKey) -> None:
        if key in self._items:
            del self._items[key]
        else:
            raise self._error_class(f"Item not found: {key}")

    def __iter__(self) -> Iterator[TKey]:
        return iter(self._items)

    async def __aiter__(self) -> AsyncIterator[tuple[TKey, TItem]]:
        """Async iterate over items, ensuring they're initialized."""
        if not self._initialized:
            await self.startup()
        for key, item in self._items.items():
            yield key, item

    def __len__(self) -> int:
        return len(self._items)



"""ACP agent implementation."""

from .acp_agent import ACPAgent
from .acp_converters import ACPMessageAccumulator, acp_notifications_to_messages

__all__ = ["ACPAgent", "ACPMessageAccumulator", "acp_notifications_to_messages"]



"""User interaction tool for asking clarifying questions."""

from __future__ import annotations

from typing import Literal

from agentpool.tool_impls.question.tool import QuestionTool
from agentpool_config.tools import ToolHints


__all__ = ["QuestionTool", "create_question_tool"]

# Tool metadata defaults
NAME = "question"
DESCRIPTION = "Ask the user a clarifying question during processing."
CATEGORY: Literal["other"] = "other"
HINTS = ToolHints(open_world=True)


def create_question_tool(
    *,
    name: str = NAME,
    description: str = DESCRIPTION,
    requires_confirmation: bool = False,
) -> QuestionTool:
    """Create a configured QuestionTool instance.

    Args:
        name: Tool name override.
        description: Tool description override.
        requires_confirmation: Whether tool execution needs confirmation.

    Returns:
        Configured QuestionTool instance.
    """
    return QuestionTool(
        name=name,
        description=description,
        category=CATEGORY,
        hints=HINTS,
        requires_confirmation=requires_confirmation,
    )
